"""Number formatting using Babel.

Ports ``_formatNum`` from the JS core library.
"""

from __future__ import annotations

from babel.numbers import format_compact_decimal, format_decimal, format_percent

from generaltranslation.formatting._helpers import _resolve_babel_locale


def format_num(
    value: int | float,
    locales: str | list[str] | None = None,
    options: dict | None = None,
) -> str:
    """Format a number according to the given locale and options.

    Args:
        value: The number to format.
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Formatting options (snake_case).

            - ``style``: ``"decimal"`` (default), ``"percent"``.
            - ``notation``: ``"standard"`` (default), ``"compact"``.
            - ``use_grouping``: Whether to use grouping separators (default ``True``).
            - ``minimum_fraction_digits``: Minimum number of fraction digits.
            - ``maximum_fraction_digits``: Maximum number of fraction digits.
            - ``minimum_integer_digits``: Minimum number of integer digits.

    Returns:
        The formatted number string.
    """
    if options is None:
        options = {}

    locale = _resolve_babel_locale(locales)
    style = options.get("style", "decimal")
    notation = options.get("notation", "standard")

    # Compact notation
    if notation == "compact":
        try:
            result = format_compact_decimal(value, locale=locale)
            return str(result)
        except Exception:
            pass

    # Percent style
    if style == "percent":
        fmt = _build_format_string(options)
        return format_percent(value, format=fmt or None, locale=locale)

    # Default: decimal
    fmt = _build_format_string(options)
    use_grouping = options.get("use_grouping", True)
    return format_decimal(
        value,
        format=fmt or None,
        locale=locale,
        group_separator=use_grouping,
    )


def _build_format_string(options: dict) -> str | None:
    """Build a Babel number format pattern string from options.

    Maps ``minimum_fraction_digits``, ``maximum_fraction_digits``, and
    ``minimum_integer_digits`` to a Babel-compatible format pattern.

    Returns:
        A pattern string like ``"#,##0.00"`` or ``None`` to use defaults.
    """
    min_frac = options.get("minimum_fraction_digits")
    max_frac = options.get("maximum_fraction_digits")
    min_int = options.get("minimum_integer_digits")

    if min_frac is None and max_frac is None and min_int is None:
        return None

    # Integer part
    int_digits = min_int or 1
    if options.get("use_grouping", True):
        # Build grouped pattern: e.g. #,##0, #,##00
        if int_digits <= 1:
            int_part = "#,##0"
        else:
            int_part = "#,##" + "0" * int_digits
    else:
        int_part = "0" * int_digits

    # Fraction part
    if min_frac is not None or max_frac is not None:
        required = min_frac or 0
        optional = (max_frac or required) - required
        frac_part = "0" * required + "#" * optional
        return f"{int_part}.{frac_part}" if frac_part else int_part

    return int_part



"""Query file data endpoint."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._request import api_request


async def query_file_data(
    data: dict[str, Any],
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Query data about one or more source or translation files."""
    body: dict[str, Any] = {}
    if data.get("source_files") or data.get("sourceFiles"):
        source_files = data.get("source_files") or data.get("sourceFiles") or []
        body["sourceFiles"] = [
            {
                "fileId": item.get("file_id", item.get("fileId", "")),
                "versionId": item.get("version_id", item.get("versionId", "")),
                "branchId": item.get("branch_id", item.get("branchId", "")),
            }
            for item in source_files
        ]
    if data.get("translated_files") or data.get("translatedFiles"):
        translated_files = data.get("translated_files") or data.get("translatedFiles") or []
        body["translatedFiles"] = [
            {
                "fileId": item.get("file_id", item.get("fileId", "")),
                "versionId": item.get("version_id", item.get("versionId", "")),
                "branchId": item.get("branch_id", item.get("branchId", "")),
                "locale": item.get("locale", ""),
            }
            for item in translated_files
        ]

    return await api_request(
        config,
        "/v2/project/files/info",
        body=body,
        timeout=options.get("timeout"),
    )



"""Determine whether translation is needed between two locales.

Ports ``requiresTranslation.ts`` from the JS core library.
"""

from __future__ import annotations

from generaltranslation.locales._types import CustomMapping


def requires_translation(
    source_locale: str,
    target_locale: str,
    approved_locales: list[str] | None = None,
    custom_mapping: CustomMapping | None = None,
) -> bool:
    """Return ``True`` if translating from *source_locale* to *target_locale* is needed.

    Translation is **not** required (returns ``False``) when:

    1. Either locale is invalid.
    2. Any entry in *approved_locales* is invalid.
    3. *source_locale* and *target_locale* are the same dialect.
    4. *target_locale* is not represented in *approved_locales*
       (i.e. the application does not support that language at all).

    Otherwise, translation is required.

    Args:
        source_locale: The locale of the source content.
        target_locale: The locale to translate into.
        approved_locales: Optional list of locales the application
            supports.  If provided and *target_locale*'s language is
            not among them, translation is skipped.
        custom_mapping: Optional custom mapping passed through to
            locale validation.

    Returns:
        ``True`` if a translation should be performed.

    Examples:
        >>> requires_translation("en", "fr")
        True
        >>> requires_translation("en-US", "en")
        False
        >>> requires_translation("en", "ja", approved_locales=["fr", "de"])
        False
    """
    from generaltranslation.locales._is_same_dialect import is_same_dialect
    from generaltranslation.locales._is_same_language import is_same_language
    from generaltranslation.locales._is_valid_locale import is_valid_locale

    # Validate both locales
    if not is_valid_locale(source_locale, custom_mapping):
        return False
    if not is_valid_locale(target_locale, custom_mapping):
        return False

    # Same dialect means no translation needed
    if is_same_dialect(source_locale, target_locale):
        return False

    # Check approved locales
    if approved_locales is not None:
        # Validate all approved locales
        for approved in approved_locales:
            if not is_valid_locale(approved, custom_mapping):
                return False

        # Check if target language is represented in approved locales
        target_represented = any(is_same_language(target_locale, approved) for approved in approved_locales)
        if not target_represented:
            return False

    return True



from __future__ import annotations

import re
from collections.abc import Callable
from typing import Any

from generaltranslation_icu_messageformat_parser import Parser


def traverse_icu(
    icu_string: str,
    should_visit: Callable[[dict], bool],
    visitor: Callable[[dict], None],
    *,
    recurse_into_visited: bool = True,
    include_indices: bool = False,
    preserve_whitespace: bool = False,
) -> list:
    """Parse an ICU string and traverse the AST, calling *visitor* on matching nodes.

    Args:
        icu_string: The ICU MessageFormat string to traverse.
        should_visit: Predicate returning ``True`` for nodes to visit.
        visitor: Called for each node where *should_visit* returns ``True``.
        recurse_into_visited: If ``False``, skip recursing into children of visited nodes.
        include_indices: If ``True``, include ``start``/``end`` on AST nodes.

    Returns:
        The (possibly mutated) AST.
    """
    parser = Parser(
        {
            "include_indices": include_indices,
            "require_other": False,
            "preserve_whitespace": preserve_whitespace,
        }
    )
    ast = parser.parse(icu_string)

    def handle_children(children: list) -> None:
        for child in children:
            handle_child(child)

    def handle_child(child: str | dict[str, Any]) -> None:
        if isinstance(child, str):
            return

        visited = False
        if should_visit(child):
            visitor(child)
            visited = True

        if not visited or recurse_into_visited:
            node_type = child.get("type")
            if node_type in ("select", "plural", "selectordinal"):
                for key, value in child.get("options", {}).items():
                    if key == "_ws":
                        continue
                    handle_children(value)
            elif node_type == "tag":
                handle_children(child.get("contents", []))

    handle_children(ast)
    return ast


_GT_INDEXED_RE = re.compile(r"^_gt_\d+$")


def is_gt_indexed_select(node: dict) -> bool:
    """Return ``True`` if *node* is a ``_gt_<number>`` select element."""
    return (
        node.get("type") == "select"
        and bool(_GT_INDEXED_RE.match(node.get("name", "")))
        and "other" in node.get("options", {})
        and (len(node["options"]["other"]) == 0 or isinstance(node["options"]["other"][0], str))
    )


def is_gt_unindexed_select(node: dict) -> bool:
    """Return ``True`` if *node* is an exact ``_gt_`` select element."""
    return (
        node.get("type") == "select"
        and node.get("name", "") == "_gt_"
        and "other" in node.get("options", {})
        and (len(node["options"]["other"]) == 0 or isinstance(node["options"]["other"][0], str))
    )



from __future__ import annotations

from generaltranslation.static._constants import VAR_IDENTIFIER, VAR_NAME_IDENTIFIER
from generaltranslation.static._sanitize_var import sanitize_var


def declare_var(
    variable: str | int | float | bool | None,
    *,
    name: str | None = None,
) -> str:
    """Mark a value as a non-translatable variable in an ICU string.

    Args:
        variable: The value to embed.
        name: Optional human-readable variable name.

    Returns:
        An ICU ``{_gt_, select, other {…}}`` construct.
    """
    if variable is None:
        raw = ""
    elif isinstance(variable, bool):
        raw = str(variable).lower()
    else:
        raw = str(variable)
    sanitized_variable = sanitize_var(raw)
    variable_section = f" other {{{sanitized_variable}}}"

    name_section = ""
    if name:
        sanitized_name = sanitize_var(name)
        name_section = f" {VAR_NAME_IDENTIFIER} {{{sanitized_name}}}"

    return f"{{{VAR_IDENTIFIER}, select,{variable_section}{name_section}}}"



from __future__ import annotations

from generaltranslation_icu_messageformat_parser import print_ast

from generaltranslation.static._constants import VAR_IDENTIFIER
from generaltranslation.static._traverse_icu import is_gt_indexed_select, traverse_icu


def condense_vars(icu_string: str) -> str:
    """Condense indexed ``_gt_`` selects to simple variable references.

    Transforms ``{_gt_1, select, other {value}}`` into ``{_gt_1}``.

    Args:
        icu_string: An ICU MessageFormat string with indexed GT variables.

    Returns:
        The string with indexed GT selects collapsed to argument references.
    """
    if VAR_IDENTIFIER not in icu_string:
        return icu_string

    def visitor(child: dict) -> None:
        # Mutate select node into a simple variable node
        child.pop("type", None)
        child.pop("options", None)
        child.pop("offset", None)

    ast = traverse_icu(
        icu_string,
        is_gt_indexed_select,
        visitor,
        recurse_into_visited=False,
    )

    return print_ast(ast)



"""TypedDicts for all translate request/response shapes."""

from __future__ import annotations

from typing import Any, Literal, TypedDict

# --- Config ---


class TranslationRequestConfig(TypedDict, total=False):
    project_id: str
    base_url: str
    api_key: str


# --- File types ---

FileFormat = Literal["GTJSON", "JSON", "YAML", "MDX", "MD", "TS", "JS", "HTML", "TXT"]
DataFormat = Literal["JSX", "ICU", "I18NEXT", "STRING"]


class FileReference(TypedDict, total=False):
    file_id: str
    version_id: str
    branch_id: str
    file_name: str
    file_format: FileFormat
    data_format: DataFormat


class FileUpload(TypedDict, total=False):
    branch_id: str
    incoming_branch_id: str
    checked_out_branch_id: str
    content: str
    file_name: str
    file_format: FileFormat
    data_format: DataFormat
    locale: str
    format_metadata: dict[str, Any]
    version_id: str
    file_id: str


# --- Translate ---

ActionType = Literal["fast"]


class EntryMetadata(TypedDict, total=False):
    id: str
    hash: str
    context: str
    max_chars: int
    data_format: DataFormat
    action_type: ActionType


class TranslateOptions(TypedDict, total=False):
    target_locale: str
    source_locale: str
    model_provider: str


class TranslationResult(TypedDict, total=False):
    success: bool
    translation: Any
    locale: str
    error: str
    code: int
    reference: dict[str, str]


class TranslationError(TypedDict):
    success: bool
    error: str
    code: int


# --- Batch ---


class BatchList(TypedDict):
    data: list[Any]
    count: int
    batch_count: int


# --- Enqueue ---


class EnqueueOptions(TypedDict, total=False):
    source_locale: str
    target_locales: list[str]
    publish: bool
    require_approval: bool
    model_provider: str
    force: bool
    timeout: int


class EnqueueFilesResult(TypedDict, total=False):
    job_data: dict[str, Any]
    locales: list[str]
    message: str


# --- Setup ---


class SetupProjectOptions(TypedDict, total=False):
    force: bool
    locales: list[str]
    timeout_ms: int


class SetupProjectResult(TypedDict, total=False):
    setup_job_id: str
    status: str


# --- Job Status ---

JobStatus = Literal["queued", "processing", "completed", "failed", "unknown"]


class JobStatusEntry(TypedDict, total=False):
    job_id: str
    status: JobStatus
    error: dict[str, str]


CheckJobStatusResult = list[JobStatusEntry]


# --- Branch ---


class BranchQuery(TypedDict):
    branch_names: list[str]


class BranchInfo(TypedDict):
    id: str
    name: str


class BranchDataResult(TypedDict, total=False):
    branches: list[BranchInfo]
    default_branch: BranchInfo | None


class CreateBranchQuery(TypedDict):
    branch_name: str
    default_branch: bool


class CreateBranchResult(TypedDict):
    branch: BranchInfo


# --- File Data ---


class FileDataQuery(TypedDict, total=False):
    source_files: list[dict[str, str]]
    translated_files: list[dict[str, str]]


class FileDataResult(TypedDict, total=False):
    source_files: list[dict[str, Any]]
    translated_files: list[dict[str, Any]]


class FileQuery(TypedDict, total=False):
    file_id: str
    branch_id: str
    version_id: str


class FileQueryResult(TypedDict, total=False):
    source_file: dict[str, Any]
    translations: list[dict[str, Any]]


class CheckFileTranslationsOptions(TypedDict, total=False):
    timeout: int


# --- Download ---


class DownloadFileBatchOptions(TypedDict, total=False):
    timeout: int


class DownloadedFile(TypedDict, total=False):
    id: str
    branch_id: str
    file_id: str
    version_id: str
    locale: str
    file_name: str
    data: str
    metadata: dict[str, Any]
    file_format: FileFormat


class DownloadFileBatchResult(TypedDict):
    files: list[DownloadedFile]
    count: int


# --- Upload ---


class UploadFilesOptions(TypedDict, total=False):
    source_locale: str
    model_provider: str
    timeout: int


class UploadFilesResponse(TypedDict, total=False):
    uploaded_files: list[FileReference]
    count: int
    message: str


# --- Orphaned Files ---


class OrphanedFile(TypedDict):
    file_id: str
    version_id: str
    file_name: str


class GetOrphanedFilesResult(TypedDict):
    orphaned_files: list[OrphanedFile]


# --- Process Moves ---


class MoveMapping(TypedDict):
    old_file_id: str
    new_file_id: str
    new_file_name: str


class MoveResult(TypedDict, total=False):
    old_file_id: str
    new_file_id: str
    success: bool
    new_source_file_id: str
    cloned_translations_count: int
    error: str


class ProcessMovesResponse(TypedDict):
    results: list[MoveResult]
    summary: dict[str, int]


class ProcessMovesOptions(TypedDict, total=False):
    timeout: int
    branch_id: str


# --- Submit User Edit Diffs ---


class SubmitUserEditDiff(TypedDict):
    file_name: str
    locale: str
    diff: str
    branch_id: str
    version_id: str
    file_id: str
    local_content: str


class SubmitUserEditDiffsPayload(TypedDict):
    diffs: list[SubmitUserEditDiff]


# --- Project Data ---


class ProjectData(TypedDict):
    id: str
    name: str
    org_id: str
    default_locale: str
    current_locales: list[str]



"""Core API request function with retry logic."""

from __future__ import annotations

import asyncio
import json
from typing import Any, Literal

import httpx

from generaltranslation._settings import DEFAULT_BASE_URL, DEFAULT_TIMEOUT
from generaltranslation.errors import (
    ApiError,
    api_error_message,
    translation_request_failed_error,
    translation_timeout_error,
)
from generaltranslation.translate._headers import generate_request_headers

MAX_RETRIES = 3
INITIAL_DELAY_MS = 500

RetryPolicy = Literal["exponential", "linear", "none"]


def _get_retry_delay(policy: RetryPolicy, attempt: int) -> float:
    """Return delay in seconds for the given retry policy and attempt."""
    if policy == "linear":
        return (INITIAL_DELAY_MS * (attempt + 1)) / 1000.0
    elif policy == "exponential":
        return (INITIAL_DELAY_MS * (2**attempt)) / 1000.0
    return 0


async def api_request(
    config: dict[str, Any],
    endpoint: str,
    *,
    body: Any = None,
    timeout: int | None = None,
    method: str = "POST",
    retry_policy: RetryPolicy = "exponential",
) -> Any:
    """Make an API request with automatic retry for 5XX errors.

    Parameters
    ----------
    config : dict
        Must contain ``project_id``; optionally ``base_url`` and ``api_key``.
    endpoint : str
        API endpoint path, e.g. ``/v2/project/jobs/info``.
    body : Any, optional
        JSON-serialisable request body.
    timeout : int, optional
        Timeout in milliseconds.  Falls back to ``DEFAULT_TIMEOUT``.
    method : str
        HTTP method (``"GET"`` or ``"POST"``).
    retry_policy : str
        ``"exponential"``, ``"linear"``, or ``"none"``.

    Returns
    -------
    Any
        Parsed JSON response.

    Raises
    ------
    ApiError
        On 4XX responses.
    Exception
        On network failures after retries are exhausted.
    """
    timeout_ms = timeout if timeout is not None else DEFAULT_TIMEOUT
    timeout_s = timeout_ms / 1000.0
    url = f"{config.get('base_url') or DEFAULT_BASE_URL}{endpoint}"
    max_retries = 0 if retry_policy == "none" else MAX_RETRIES
    headers = generate_request_headers(config)

    async with httpx.AsyncClient(timeout=timeout_s) as client:
        for attempt in range(max_retries + 1):
            try:
                if method == "GET":
                    response = await client.get(url, headers=headers)
                else:
                    response = await client.post(
                        url,
                        headers=headers,
                        content=json.dumps(body) if body is not None else None,
                    )
            except httpx.TimeoutException:
                if attempt < max_retries:
                    await asyncio.sleep(_get_retry_delay(retry_policy, attempt))
                    continue
                raise Exception(translation_timeout_error(timeout_ms))
            except Exception as exc:
                if attempt < max_retries:
                    await asyncio.sleep(_get_retry_delay(retry_policy, attempt))
                    continue
                raise Exception(translation_request_failed_error(str(exc))) from exc

            # Retry on 5XX server errors
            if response.status_code >= 500 and attempt < max_retries:
                await asyncio.sleep(_get_retry_delay(retry_policy, attempt))
                continue

            # Validate response
            if response.status_code >= 400:
                error_msg = "Unknown error"
                try:
                    text = response.text
                    try:
                        error_json = json.loads(text)
                        error_msg = error_json.get("error", text or "Unknown error")
                    except (json.JSONDecodeError, TypeError):
                        error_msg = text or "Unknown error"
                except Exception:
                    pass
                error_message = api_error_message(response.status_code, response.reason_phrase or "", error_msg)
                raise ApiError(error_message, response.status_code, error_msg)

            return response.json()

        raise Exception("Max retries exceeded")



"""Plural form selection for a given number.

Ports ``getPluralForm.ts`` from the JS core library.
Uses CLDR plural rules (via ``babel``) instead of the browser
``Intl.PluralRules`` API.
"""

from __future__ import annotations

from typing import cast

from babel import Locale

from generaltranslation._settings import LIBRARY_DEFAULT_LOCALE
from generaltranslation.locales._types import PLURAL_FORMS, PluralType

# Aliases between our custom form names and CLDR form names
_ALIASES: dict[str, str] = {
    "one": "singular",
    "singular": "one",
    "two": "dual",
    "dual": "two",
    "other": "plural",
    "plural": "other",
}


def get_plural_form(
    n: int | float,
    forms: list[PluralType] | None = None,
    locales: list[str] | None = None,
) -> PluralType | str:
    """Determine the plural form for number *n*.

    Selects the most appropriate plural branch from *forms* based on
    CLDR plural rules for the given *locales*.

    The selection follows a priority system with overrides:

    - ``n == 0`` and ``"zero"`` is available -> ``"zero"``
    - ``|n| == 1`` and ``"singular"``/``"one"`` is available -> use it
    - ``|n| == 2`` and ``"dual"``/``"two"`` is available -> use it
    - Otherwise, consult CLDR plural rules for the locale and find the
      best available form, with aliases.

    Args:
        n: The number to select a plural form for.
        forms: The allowed plural forms to choose from.
        locales: BCP 47 locale tags for plural rule selection.
            Defaults to ``["en"]``.

    Returns:
        The selected plural form name, or ``""`` if no form fits.
    """
    if forms is None:
        forms = list(PLURAL_FORMS)
    if locales is None:
        locales = [LIBRARY_DEFAULT_LOCALE]

    forms_set = set(forms)

    # Override: n == 0 and "zero" is available
    if n == 0 and "zero" in forms_set:
        return "zero"

    # Override: |n| == 1 and "singular" or "one" is available
    if abs(n) == 1:
        if "singular" in forms_set:
            return "singular"
        if "one" in forms_set:
            return "one"

    # Override: |n| == 2 and "dual" or "two" is available
    if abs(n) == 2:
        if "dual" in forms_set:
            return "dual"
        if "two" in forms_set:
            return "two"

    # Get CLDR plural category for the locale
    cldr_category = _get_cldr_category(n, locales)

    # Try to find the best matching form
    return _find_best_form(cldr_category, cast("set[str]", forms_set), cast("list[str]", forms))


def _get_cldr_category(n: int | float, locales: list[str]) -> str:
    """Get CLDR plural category for n in the given locale."""
    for locale_tag in locales:
        try:
            parsed = Locale.parse(locale_tag, sep="-")
            rule = parsed.plural_form
            category = rule(abs(n))
            return category
        except Exception:
            continue
    # Default to English rules
    try:
        rule = Locale(LIBRARY_DEFAULT_LOCALE).plural_form
        return rule(abs(n))
    except Exception:
        return "other"


def _find_best_form(
    cldr_category: str,
    forms_set: set[str],
    forms: list[str],
) -> str:
    """Find the best available form matching the CLDR category."""
    # Direct match
    if cldr_category in forms_set:
        return cldr_category

    # Try alias
    alias = _ALIASES.get(cldr_category)
    if alias and alias in forms_set:
        return alias

    # Fallback to "other" or "plural"
    if "other" in forms_set:
        return "other"
    if "plural" in forms_set:
        return "plural"

    # Last resort: return the last form in the list
    if forms:
        return forms[-1]

    return ""



"""ICU MessageFormat formatting.

Delegates to ``generaltranslation-intl-messageformat`` for parsing and
locale-aware interpolation (plural, select, selectordinal).
"""

from __future__ import annotations

from generaltranslation_intl_messageformat import IntlMessageFormat

from generaltranslation.formatting._helpers import _resolve_babel_locale


def format_message(
    message: str,
    locales: str | list[str] | None = None,
    variables: dict | None = None,
) -> str:
    """Format an ICU MessageFormat string.

    Args:
        message: The ICU message pattern (e.g.
            ``"{count, plural, one {# item} other {# items}}"``).
        locales: BCP 47 locale tag(s) for plural rule selection.
            Defaults to ``"en"``.
        variables: Variable values to substitute.

    Returns:
        The formatted message string.
    """
    locale = _resolve_babel_locale(locales)
    mf = IntlMessageFormat(message, str(locale))
    return mf.format(variables)



"""Upload translations endpoint."""

from __future__ import annotations

import base64
from typing import Any

from generaltranslation.translate._batch import process_batches
from generaltranslation.translate._request import api_request


async def upload_translations(
    files: list[dict[str, Any]],
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Upload translations to the API in batches."""

    async def _process_batch(batch: list[dict[str, Any]]) -> list[Any]:
        body = {
            "data": [
                {
                    "source": {
                        "content": base64.b64encode(item["source"]["content"].encode()).decode(),
                        "fileName": item["source"].get("file_name", item["source"].get("fileName", "")),
                        "fileFormat": item["source"].get("file_format", item["source"].get("fileFormat", "")),
                        "locale": item["source"].get("locale", ""),
                        "dataFormat": item["source"].get("data_format", item["source"].get("dataFormat")),
                        "formatMetadata": item["source"].get("format_metadata", item["source"].get("formatMetadata")),
                        "fileId": item["source"].get("file_id", item["source"].get("fileId")),
                        "versionId": item["source"].get("version_id", item["source"].get("versionId")),
                        "branchId": item["source"].get("branch_id", item["source"].get("branchId")),
                    },
                    "translations": [
                        {
                            "content": base64.b64encode(t["content"].encode()).decode(),
                            "fileName": t.get("file_name", t.get("fileName", "")),
                            "fileFormat": t.get("file_format", t.get("fileFormat", "")),
                            "locale": t.get("locale", ""),
                            "dataFormat": t.get("data_format", t.get("dataFormat")),
                            "fileId": t.get("file_id", t.get("fileId")),
                            "versionId": t.get("version_id", t.get("versionId")),
                            "branchId": t.get("branch_id", t.get("branchId")),
                        }
                        for t in item.get("translations", [])
                    ],
                }
                for item in batch
            ],
            "sourceLocale": options.get("source_locale", options.get("sourceLocale", "")),
        }
        result = await api_request(
            config,
            "/v2/project/files/upload-translations",
            body=body,
            timeout=options.get("timeout"),
        )
        return result.get("uploadedFiles", [])

    return await process_batches(files, _process_batch, batch_size=100)



"""Get project data endpoint."""

from __future__ import annotations

from typing import Any
from urllib.parse import quote

from generaltranslation.translate._request import api_request


async def get_project_data(
    project_id: str,
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Get project data for a given project ID."""
    encoded_id = quote(project_id, safe="!'()*-._~")
    endpoint = f"/v2/project/info/{encoded_id}"
    return await api_request(
        config,
        endpoint,
        method="GET",
        timeout=options.get("timeout"),
    )



GT_ERROR_PREFIX = "GT Error:"


def translation_timeout_error(timeout: int) -> str:
    return f"{GT_ERROR_PREFIX} Translation request timed out after {timeout}ms."


def translation_request_failed_error(error: str) -> str:
    return f"{GT_ERROR_PREFIX} Translation request failed. Error: {error}"


def api_error_message(status: int, status_text: str, error: str) -> str:
    return f"{GT_ERROR_PREFIX} API returned error status. Status: {status}, Status Text: {status_text}, Error: {error}"


INVALID_AUTH_ERROR = f"{GT_ERROR_PREFIX} Invalid authentication."


def no_target_locale_error(fn_name: str) -> str:
    return (
        f"{GT_ERROR_PREFIX} Cannot call `{fn_name}` without a specified locale."
        f" Either pass a locale to the `{fn_name}` function or specify a targetLocale in the GT constructor."
    )


def no_source_locale_error(fn_name: str) -> str:
    return (
        f"{GT_ERROR_PREFIX} Cannot call `{fn_name}` without a specified locale."
        f" Either pass a locale to the `{fn_name}` function or specify a sourceLocale in the GT constructor."
    )


def no_project_id_error(fn_name: str) -> str:
    return (
        f"{GT_ERROR_PREFIX} Cannot call `{fn_name}` without a specified project ID."
        f" Either pass a project ID to the `{fn_name}` function or specify a projectId in the GT constructor."
    )


def no_api_key_error(fn_name: str) -> str:
    return (
        f"{GT_ERROR_PREFIX} Cannot call `{fn_name}` without a specified API key."
        f" Either pass an API key to the `{fn_name}` function or specify an apiKey in the GT constructor."
    )


def invalid_locale_error(locale: str) -> str:
    return f"{GT_ERROR_PREFIX} Invalid locale: {locale}."


def invalid_locales_error(locales: list[str]) -> str:
    return f"{GT_ERROR_PREFIX} Invalid locales: {', '.join(locales)}."


def create_invalid_cutoff_style_error(style: str) -> str:
    return f"generaltranslation Formatting Error: Invalid cutoff style: {style}."



"""Download multiple translation files in batches."""

from __future__ import annotations

import base64
from typing import Any

from generaltranslation.translate._batch import process_batches
from generaltranslation.translate._request import api_request


async def download_file_batch(
    requests: list[dict[str, Any]],
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Download multiple translation files in batches."""

    async def _process_batch(batch: list[dict[str, Any]]) -> list[Any]:
        result = await api_request(
            config,
            "/v2/project/files/download",
            body=batch,
            timeout=options.get("timeout"),
        )
        files = []
        for f in result.get("files", []):
            decoded = {**f}
            decoded["data"] = base64.b64decode(f["data"]).decode("utf-8")
            files.append(decoded)
        return files

    return await process_batches(requests, _process_batch, batch_size=100)



from __future__ import annotations

from generaltranslation.static._constants import VAR_IDENTIFIER
from generaltranslation.static._traverse_icu import is_gt_unindexed_select, traverse_icu


def extract_vars(icu_string: str) -> dict[str, str]:
    """Extract ``_gt_`` variables from an ICU string into an indexed mapping.

    Args:
        icu_string: An ICU MessageFormat string with unindexed GT variables.

    Returns:
        A dict mapping ``_gt_1``, ``_gt_2``, … to their ``other`` option values.
    """
    if VAR_IDENTIFIER not in icu_string:
        return {}

    variables: dict[str, str] = {}
    index = [1]

    def visitor(child: dict) -> None:
        other = child["options"]["other"]
        variables[child["name"] + str(index[0])] = other[0] if other else ""
        index[0] += 1

    traverse_icu(
        icu_string,
        is_gt_unindexed_select,
        visitor,
        recurse_into_visited=False,
    )

    return variables



from __future__ import annotations


def declare_static(content: object) -> object:
    """Mark *content* as statically analyzable.

    This is an identity function used as a marker for static analysis.
    """
    return content



"""Language-level locale comparison.

Ports ``isSameLanguage.ts`` from the JS core library.
Compares only the language subtag, ignoring region and script.
"""

from __future__ import annotations

from babel import Locale


def is_same_language(*locales: str | list[str]) -> bool:
    """Return ``True`` if all given locales share the same base language.

    Extracts the language component from each BCP 47 tag (ignoring
    region and script) and checks that they are all identical.

    Accepts individual strings or lists of strings (which are flattened).

    Args:
        *locales: Two or more BCP 47 locale tags (or lists thereof).

    Returns:
        ``True`` if every locale has the same language code,
        ``False`` otherwise (including on parse errors).

    Examples:
        >>> is_same_language("en-US", "en-GB")
        True
        >>> is_same_language("en-US", "fr-FR")
        False
        >>> is_same_language("zh-Hans", "zh-Hant")
        True
    """
    # Flatten args
    flat: list[str] = []
    for arg in locales:
        if isinstance(arg, list):
            flat.extend(arg)
        else:
            flat.append(arg)

    if len(flat) < 2:
        return False

    try:
        languages: list[str] = []
        for loc in flat:
            parsed = Locale.parse(loc, sep="-")
            languages.append(parsed.language)
        return len(set(languages)) == 1
    except Exception:
        return False



"""Locale hierarchy / superset checking.

Ports ``isSupersetLocale.ts`` from the JS core library.
A locale is a "superset" of another if it is equal to or more general
than the other (e.g. ``"en"`` is a superset of ``"en-US"``).
"""

from __future__ import annotations

from babel import Locale


def is_superset_locale(super_locale: str, sub_locale: str) -> bool:
    """Return ``True`` if *super_locale* is a superset of *sub_locale*.

    *super_locale* is a superset when:

    - Both have the same language code.
    - *super_locale* either has no region or its region matches
      *sub_locale*'s region.
    - *super_locale* either has no script or its script matches
      *sub_locale*'s script.

    In other words, a more general locale (fewer subtags) is always a
    superset of a more specific one with the same language.

    Args:
        super_locale: The potentially more general locale.
        sub_locale: The potentially more specific locale.

    Returns:
        ``True`` if *super_locale* encompasses *sub_locale*,
        ``False`` otherwise (including on parse errors).

    Examples:
        >>> is_superset_locale("en", "en-US")
        True
        >>> is_superset_locale("en-US", "en")
        False
        >>> is_superset_locale("en-US", "en-US")
        True
        >>> is_superset_locale("en-US", "en-GB")
        False
    """
    try:
        from generaltranslation.locales._is_valid_locale import standardize_locale

        sup = Locale.parse(standardize_locale(super_locale), sep="-")
        sub = Locale.parse(standardize_locale(sub_locale), sep="-")

        if sup.language != sub.language:
            return False

        if sup.territory and sup.territory != sub.territory:
            return False

        if sup.script and sup.script != sub.script:
            return False

        # super can't be more specific than sub
        if sup.territory and not sub.territory:
            return False

        if sup.script and not sub.script:
            return False

        return True
    except Exception:
        return False



"""Number, currency, datetime, and list formatting."""

from generaltranslation.formatting._format_currency import format_currency
from generaltranslation.formatting._format_cutoff import CutoffFormat, format_cutoff
from generaltranslation.formatting._format_date_time import format_date_time
from generaltranslation.formatting._format_list import format_list, format_list_to_parts
from generaltranslation.formatting._format_message import format_message
from generaltranslation.formatting._format_num import format_num
from generaltranslation.formatting._format_relative_time import format_relative_time

__all__ = [
    "format_num",
    "format_currency",
    "format_date_time",
    "format_list",
    "format_list_to_parts",
    "format_relative_time",
    "format_message",
    "format_cutoff",
    "CutoffFormat",
]



"""Canonical ↔ alias locale resolution.

Ports ``resolveCanonicalLocale.ts`` and ``resolveAliasLocale.ts`` from
the JS core library.  These two functions form an invertible mapping
between user-defined locale aliases and their canonical BCP 47 codes.
"""

from __future__ import annotations

from generaltranslation.locales._types import CustomMapping


def resolve_canonical_locale(
    locale: str,
    custom_mapping: CustomMapping | None = None,
) -> str:
    """Resolve a locale alias to its canonical BCP 47 code.

    If *custom_mapping* contains an entry for *locale* that is a dict
    with a ``"code"`` key pointing to a valid locale, that code is
    returned.  Otherwise *locale* is returned unchanged.

    Args:
        locale: A locale tag or alias to resolve.
        custom_mapping: Optional custom mapping defining aliases.

    Returns:
        The canonical locale code.

    Examples:
        >>> mapping = {"pt-custom": {"code": "pt-BR", "name": "Custom PT"}}
        >>> resolve_canonical_locale("pt-custom", mapping)
        'pt-BR'
        >>> resolve_canonical_locale("en")
        'en'
    """
    if custom_mapping is not None:
        from generaltranslation.locales._custom_locale_mapping import (
            should_use_canonical_locale,
        )

        if should_use_canonical_locale(locale, custom_mapping):
            entry = custom_mapping[locale]
            if isinstance(entry, dict):
                return entry["code"]
    return locale


def resolve_alias_locale(
    locale: str,
    custom_mapping: CustomMapping | None = None,
) -> str:
    """Resolve a canonical locale code back to its alias.

    Builds a reverse index from *custom_mapping* (canonical code →
    alias key) and returns the alias for *locale* if one exists.
    Otherwise returns *locale* unchanged.

    Args:
        locale: A canonical BCP 47 locale code.
        custom_mapping: Optional custom mapping defining aliases.

    Returns:
        The alias locale string, or *locale* if no alias is defined.

    Examples:
        >>> mapping = {"pt-custom": {"code": "pt-BR", "name": "Custom PT"}}
        >>> resolve_alias_locale("pt-BR", mapping)
        'pt-custom'
        >>> resolve_alias_locale("en", mapping)
        'en'
    """
    if custom_mapping is not None:
        from generaltranslation.locales._custom_locale_mapping import (
            should_use_canonical_locale,
        )

        for alias, entry in custom_mapping.items():
            if should_use_canonical_locale(alias, custom_mapping):
                if isinstance(entry, dict) and entry.get("code") == locale:
                    return alias
    return locale



"""Text direction lookup for BCP 47 locales.

Ports ``getLocaleDirection.ts`` from the JS core library.
Uses ``langcodes`` script metadata to determine whether a locale's
writing system is left-to-right or right-to-left.
"""

from __future__ import annotations

from babel import Locale


def get_locale_direction(locale: str) -> str:
    """Return the text direction for *locale*.

    Determines whether the script used by *locale* is written
    left-to-right or right-to-left.  Uses ``langcodes`` to resolve the
    likely script and checks it against a known set of RTL scripts
    (Arabic, Hebrew, Thaana, Syriac, etc.).

    Args:
        locale: A BCP 47 locale tag (e.g. ``"ar"``, ``"en-US"``).

    Returns:
        ``"rtl"`` if the locale uses a right-to-left script, otherwise
        ``"ltr"``.  Defaults to ``"ltr"`` if the locale cannot be parsed.

    Examples:
        >>> get_locale_direction("ar")
        'rtl'
        >>> get_locale_direction("en-US")
        'ltr'
        >>> get_locale_direction("he")
        'rtl'
    """
    try:
        parsed = Locale.parse(locale, sep="-")
        return parsed.text_direction
    except Exception:
        return "ltr"



"""Create branch endpoint."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._request import api_request


async def create_branch(
    query: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Create a new branch in the API."""
    return await api_request(
        config,
        "/v2/project/branches/create",
        body=query,
    )



from generaltranslation.static._condense_vars import condense_vars
from generaltranslation.static._constants import VAR_IDENTIFIER, VAR_NAME_IDENTIFIER
from generaltranslation.static._declare_static import declare_static
from generaltranslation.static._declare_var import declare_var
from generaltranslation.static._decode_vars import decode_vars
from generaltranslation.static._extract_vars import extract_vars
from generaltranslation.static._index_vars import index_vars
from generaltranslation.static._sanitize_var import sanitize_var

__all__ = [
    "VAR_IDENTIFIER",
    "VAR_NAME_IDENTIFIER",
    "sanitize_var",
    "declare_var",
    "declare_static",
    "decode_vars",
    "extract_vars",
    "index_vars",
    "condense_vars",
]



"""BCP 47 locale validation and standardisation.

Ports ``isValidLocale.ts`` from the JS core library.
Uses the ``babel`` library instead of the browser ``Intl.Locale`` API.
"""

from __future__ import annotations

from babel import Locale
from babel.core import UnknownLocaleError

from generaltranslation._settings import LIBRARY_DEFAULT_LOCALE
from generaltranslation.locales._types import CustomMapping

# Scripts that are valid but may not be recognised by all display-name APIs.
SCRIPT_EXCEPTIONS: set[str] = {"Cham", "Jamo", "Kawi", "Lisu", "Toto", "Thai"}


def _is_custom_language(language: str) -> bool:
    """Check if *language* is in the BCP 47 private-use range (qaa-qtz).

    Per BCP 47, codes in this range are reserved for private use and are
    always considered valid.

    Args:
        language: A 2- or 3-letter ISO 639 language code (lowercase).

    Returns:
        ``True`` if the code is in the private-use range.
    """
    lang = language.lower()
    return len(lang) == 3 and lang >= "qaa" and lang <= "qtz"


def is_valid_locale(
    locale: str,
    custom_mapping: CustomMapping | None = None,
) -> bool:
    """Validate a BCP 47 language tag.

    Checks that *locale* can be parsed into valid language, region, and
    script components.  Custom mapping entries whose ``"code"`` key points
    to a valid locale are also accepted.  Private-use language codes
    (``qaa``-``qtz``) are always valid.

    Args:
        locale: The BCP 47 language tag to validate (e.g. ``"en-US"``,
            ``"zh-Hans-CN"``).
        custom_mapping: Optional custom mapping that may redirect
            *locale* to a canonical code before validation.

    Returns:
        ``True`` if the locale is valid, ``False`` otherwise.

    Examples:
        >>> is_valid_locale("en-US")
        True
        >>> is_valid_locale("not-a-locale")
        False
        >>> is_valid_locale("qaa")  # private-use range
        True
    """
    if not locale or not isinstance(locale, str):
        return False

    # Check custom mapping redirect
    if custom_mapping is not None:
        from generaltranslation.locales._custom_locale_mapping import (
            should_use_canonical_locale,
        )

        if should_use_canonical_locale(locale, custom_mapping):
            entry = custom_mapping[locale]
            if isinstance(entry, dict):
                return is_valid_locale(entry["code"])

    # Split using hyphen separator
    sep = "-"
    parts = locale.split(sep)
    if len(parts) > 3:
        return False

    language = parts[0].lower()

    # Private-use range is always valid
    if _is_custom_language(language):
        return True

    try:
        parsed = Locale.parse(locale, sep=sep)
    except (UnknownLocaleError, ValueError):
        return False

    # Validate that parsed components look reasonable
    lang = parsed.language
    if not lang or len(lang) < 2:
        return False

    # Verify Babel didn't silently drop input parts (e.g. "en-FAKE")
    # Babel may ADD parts via likely subtag resolution (zh-CN → zh-Hans-CN)
    # but should never have fewer recognized parts than the input
    parsed_part_count = 1  # language always present
    if parsed.script:
        parsed_part_count += 1
    if parsed.territory:
        parsed_part_count += 1
    if parsed_part_count < len(parts):
        return False

    # For region codes, verify they're known
    if parsed.territory:
        try:
            display_locale = Locale(LIBRARY_DEFAULT_LOCALE)
            if parsed.territory not in display_locale.territories:
                return False
        except Exception:
            pass

    # For script codes, verify they're known (with exceptions)
    if parsed.script:
        if parsed.script not in SCRIPT_EXCEPTIONS:
            try:
                display_locale = Locale(LIBRARY_DEFAULT_LOCALE)
                if parsed.script not in display_locale.scripts:
                    return False
            except Exception:
                pass

    return True


def standardize_locale(locale: str) -> str:
    """Normalise a BCP 47 locale tag to its canonical form.

    Converts underscores to hyphens, fixes casing (e.g. ``"en_us"``
    -> ``"en-US"``), and applies IANA subtag registry canonicalisation.

    Note: Tags containing underscores are returned unchanged (matching
    JS ``Intl.getCanonicalLocales()`` behavior which rejects them).

    Args:
        locale: The locale tag to standardise.

    Returns:
        The canonical form of the locale, or the original string
        unchanged if parsing fails.

    Examples:
        >>> standardize_locale("en_us")
        'en_us'
        >>> standardize_locale("zh-hans-cn")
        'zh-Hans-CN'
    """
    if not locale or not isinstance(locale, str):
        return locale

    # Underscore inputs stay unchanged (matching JS behavior)
    if "_" in locale:
        return locale

    try:
        parsed = Locale.parse(locale, sep="-")
        # Reconstruct with hyphens
        result = parsed.language
        if parsed.script:
            result += f"-{parsed.script}"
        if parsed.territory:
            result += f"-{parsed.territory}"
        return result
    except Exception:
        return locale



"""Currency formatting using Babel.

Ports ``_formatCurrency`` from the JS core library.
"""

from __future__ import annotations

from babel.numbers import format_currency as babel_format_currency

from generaltranslation.formatting._helpers import _resolve_babel_locale

# Map JS currencyDisplay values to Babel format_type
_DISPLAY_MAP = {
    "symbol": None,  # Babel default
    "narrowSymbol": None,  # Babel doesn't distinguish; use default
    "name": "name",
}


def format_currency(
    value: int | float,
    currency: str = "USD",
    locales: str | list[str] | None = None,
    options: dict | None = None,
) -> str:
    """Format a currency value according to the given locale and options.

    Args:
        value: The currency amount.
        currency: ISO 4217 currency code (e.g. ``"USD"``, ``"EUR"``).
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Formatting options (snake_case).

            - ``currency_display``: ``"symbol"`` (default), ``"name"``,
              ``"code"``, ``"narrowSymbol"``.
            - ``currency_sign``: ``"standard"`` (default), ``"accounting"``.

    Returns:
        The formatted currency string.
    """
    if options is None:
        options = {}

    locale = _resolve_babel_locale(locales)

    currency_display = options.get("currency_display", "symbol")
    currency_sign = options.get("currency_sign", "standard")

    # Determine Babel format_type
    format_type = _DISPLAY_MAP.get(currency_display)

    # Accounting format for negative values
    if currency_sign == "accounting":
        format_type = "accounting"

    # "code" display: format with symbol then replace symbol with currency code
    if currency_display == "code":
        formatted = babel_format_currency(
            number=value,
            currency=currency.upper(),
            locale=locale,
            format_type="standard",
        )
        # Get the currency symbol to replace it with the code
        from babel.numbers import get_currency_symbol

        symbol = get_currency_symbol(currency.upper(), locale=locale)
        return formatted.replace(symbol, currency.upper())

    kwargs: dict = {
        "number": value,
        "currency": currency.upper(),
        "locale": locale,
    }
    if format_type is not None:
        kwargs["format_type"] = format_type

    return babel_format_currency(**kwargs)



"""Check job status endpoint."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._request import api_request


async def check_job_status(
    job_ids: list[str],
    config: dict[str, Any],
    timeout_ms: int | None = None,
) -> list[dict[str, Any]]:
    """Query job statuses for a project."""
    return await api_request(
        config,
        "/v2/project/jobs/info",
        body={"jobIds": job_ids},
        timeout=timeout_ms,
    )



"""Centralized constants for the generaltranslation package."""

# From settings/settings.ts
LIBRARY_DEFAULT_LOCALE = "en"
DEFAULT_TIMEOUT = 60_000  # ms

# From settings/settingsUrls.ts
DEFAULT_BASE_URL = "https://api2.gtx.dev"
DEFAULT_CACHE_URL = "https://cdn.gtx.dev"
DEFAULT_RUNTIME_API_URL = "https://runtime2.gtx.dev"

# From translate/api.ts
API_VERSION = "2026-02-18.v1"



"""Relative time formatting using Babel.

Ports ``_formatRelativeTime`` from the JS core library.
"""

from __future__ import annotations

from datetime import timedelta

from babel.dates import format_timedelta

from generaltranslation.formatting._helpers import _resolve_babel_locale

# Map JS RelativeTimeFormat units to timedelta kwargs
_UNIT_TO_TIMEDELTA: dict[str, str] = {
    "second": "seconds",
    "seconds": "seconds",
    "minute": "minutes",
    "minutes": "minutes",
    "hour": "hours",
    "hours": "hours",
    "day": "days",
    "days": "days",
    "week": "weeks",
    "weeks": "weeks",
    "month": "days",  # approximate: 30 days per month
    "months": "days",
    "year": "days",  # approximate: 365 days per year
    "years": "days",
    "quarter": "days",  # approximate: 91 days per quarter
    "quarters": "days",
}

# Multipliers for units that need conversion to days
_UNIT_MULTIPLIER: dict[str, int] = {
    "month": 30,
    "months": 30,
    "year": 365,
    "years": 365,
    "quarter": 91,
    "quarters": 91,
}

# Map JS style names to Babel granularity
_STYLE_MAP: dict[str, str] = {
    "long": "long",
    "short": "short",
    "narrow": "narrow",
}


def format_relative_time(
    value: int | float,
    unit: str,
    locales: str | list[str] | None = None,
    options: dict | None = None,
) -> str:
    """Format a relative time value (e.g. "3 days ago", "in 2 hours").

    Args:
        value: The numeric value (positive = future, negative = past).
        unit: The time unit: ``"second"``, ``"minute"``, ``"hour"``,
            ``"day"``, ``"week"``, ``"month"``, ``"quarter"``, ``"year"``.
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Formatting options (snake_case).

            - ``style``: ``"long"`` (default), ``"short"``, ``"narrow"``.
            - ``numeric``: ``"auto"`` (default), ``"always"``.

    Returns:
        The formatted relative time string.
    """
    if options is None:
        options = {}

    locale = _resolve_babel_locale(locales)
    style = options.get("style", "long")
    babel_format = _STYLE_MAP.get(style, "long")

    # Build timedelta
    td_key = _UNIT_TO_TIMEDELTA.get(unit, "seconds")
    multiplier = _UNIT_MULTIPLIER.get(unit, 1)
    delta = timedelta(**{td_key: value * multiplier})

    # Use threshold=999 to prevent Babel from auto-rounding units
    # (match JS literal behavior where "5 seconds" stays as seconds)
    return format_timedelta(
        delta,
        granularity=_singular_unit(unit),  # type: ignore[arg-type]
        threshold=999,
        add_direction=True,
        format=babel_format,  # type: ignore[arg-type]
        locale=locale,
    )


def _singular_unit(unit: str) -> str:
    """Normalize unit to singular form for Babel granularity param."""
    singular_map = {
        "seconds": "second",
        "minutes": "minute",
        "hours": "hour",
        "days": "day",
        "weeks": "week",
        "months": "month",
        "quarters": "quarter",
        "years": "year",
    }
    return singular_map.get(unit, unit)



"""Setup project endpoint."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._request import api_request


async def setup_project(
    files: list[dict[str, Any]],
    config: dict[str, Any],
    options: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Enqueue files for project setup."""
    options = options or {}
    body: dict[str, Any] = {
        "files": [
            {
                "branchId": f.get("branch_id", f.get("branchId", "")),
                "fileId": f.get("file_id", f.get("fileId", "")),
                "versionId": f.get("version_id", f.get("versionId", "")),
            }
            for f in files
        ],
    }
    if options.get("locales"):
        body["locales"] = options["locales"]
    if options.get("force") is not None:
        body["force"] = options["force"]

    return await api_request(
        config,
        "/v2/project/setup/generate",
        body=body,
        timeout=options.get("timeout_ms"),
    )



import hashlib
import json
from collections.abc import Callable
from typing import Any


def hash_string(s: str) -> str:
    """SHA256 hash, first 16 hex characters."""
    return hashlib.sha256(s.encode()).hexdigest()[:16]


def hash_source(
    source: Any,
    *,
    context: str | None = None,
    id: str | None = None,
    max_chars: int | None = None,
    data_format: str = "STRING",
    hash_function: Callable[[str], str] | None = None,
) -> str:
    """Hash source content with metadata. Only ICU/STRING path (no JSX)."""
    if hash_function is None:
        hash_function = hash_string

    sanitized_data: dict[str, Any] = {"dataFormat": data_format}
    # For non-JSX, source is used as-is (string)
    sanitized_data["source"] = source

    if id:
        sanitized_data["id"] = id
    if context:
        sanitized_data["context"] = context
    if max_chars is not None:
        sanitized_data["maxChars"] = abs(max_chars)

    stringified = json.dumps(sanitized_data, sort_keys=True, separators=(",", ":"))
    return hash_function(stringified)


def hash_template(
    template: dict[str, str],
    hash_function: Callable[[str], str] | None = None,
) -> str:
    """Hash a template dict."""
    if hash_function is None:
        hash_function = hash_string
    stringified = json.dumps(template, sort_keys=True, separators=(",", ":"))
    return hash_function(stringified)



from __future__ import annotations

from generaltranslation.static._constants import VAR_IDENTIFIER
from generaltranslation.static._traverse_icu import is_gt_unindexed_select, traverse_icu


def decode_vars(icu_string: str) -> str:
    """Replace ``{_gt_, select, other {value}}`` constructs with their values.

    Args:
        icu_string: An ICU MessageFormat string potentially containing GT variables.

    Returns:
        The string with GT variable selects replaced by their embedded values.
    """
    if VAR_IDENTIFIER not in icu_string:
        return icu_string

    variable_locations: list[tuple[int, int, str]] = []

    def visitor(child: dict) -> None:
        other = child["options"]["other"]
        value = other[0] if other else ""
        variable_locations.append((child.get("start", 0), child.get("end", 0), value))

    traverse_icu(
        icu_string,
        is_gt_unindexed_select,
        visitor,
        recurse_into_visited=False,
        include_indices=True,
    )

    parts: list[str] = []
    prev = 0
    for start, end, value in variable_locations:
        parts.append(icu_string[prev:start])
        parts.append(value)
        prev = end
    if prev < len(icu_string):
        parts.append(icu_string[prev:])

    return "".join(parts)



import re


def sanitize_var(string: str) -> str:
    """Sanitize a string so it can be safely embedded in ICU select syntax.

    Escapes ICU special characters by:
    1. Doubling all ASCII single quotes (U+0027)
    2. Wrapping the region containing special characters ({}<>) in single quotes
    """
    # Double all ASCII single quotes (U+0027 only)
    result = string.replace("'", "''")

    # Find first special character position
    special = re.compile(r"[{}<>]")
    match = special.search(result)
    if match is None:
        return result

    first_special_index = match.start()

    # Find last special character position
    last_special_index = -1
    for i in range(len(result) - 1, -1, -1):
        if result[i] in "{}<>":
            last_special_index = i
            break

    # Wrap the special character region in single quotes
    result = (
        result[:first_special_index]
        + "'"
        + result[first_special_index : last_special_index + 1]
        + "'"
        + result[last_special_index + 1 :]
    )

    return result



"""Get the human-readable display name for a BCP 47 locale.

Ports ``getLocaleName.ts`` from the JS core library.
Uses ``babel`` display-name methods instead of ``Intl.DisplayNames``.
"""

from __future__ import annotations

from babel import Locale

from generaltranslation._settings import LIBRARY_DEFAULT_LOCALE
from generaltranslation.locales._types import CustomMapping


def get_locale_name(
    locale: str,
    default_locale: str | None = LIBRARY_DEFAULT_LOCALE,
    custom_mapping: CustomMapping | None = None,
) -> str:
    """Return the display name of *locale* rendered in *default_locale*.

    Resolution order:

    1. Check *custom_mapping* for a ``"name"`` override (trying the
       aliased code, standardised code, and base language code in turn).
    2. Use ``babel`` to produce a display name in *default_locale*.
    3. Fall back to an empty string if all lookups fail.

    Args:
        locale: A BCP 47 locale tag (e.g. ``"de"``, ``"zh-Hans-CN"``).
        default_locale: The language in which the name should be
            rendered (defaults to ``"en"``).
        custom_mapping: Optional custom mapping that may override the
            display name or redirect to a canonical code.

    Returns:
        The localised display name, or ``""`` on failure.

    Examples:
        >>> get_locale_name("de")
        'German'
        >>> get_locale_name("de", default_locale="fr")
        'allemand'
    """
    from generaltranslation.locales._custom_locale_mapping import (
        get_custom_property,
        should_use_canonical_locale,
    )
    from generaltranslation.locales._is_valid_locale import standardize_locale

    if not locale:
        return ""

    # Check custom mapping
    if custom_mapping is not None:
        # Try locale as-is
        custom_name = get_custom_property(custom_mapping, locale, "name")
        if custom_name:
            return custom_name

        # If it redirects to canonical, resolve
        if should_use_canonical_locale(locale, custom_mapping):
            entry = custom_mapping[locale]
            if isinstance(entry, dict):
                canonical = entry["code"]
                custom_name = get_custom_property(custom_mapping, canonical, "name")
                if custom_name:
                    return custom_name
                locale = canonical

        # Try standardized
        std = standardize_locale(locale)
        if std != locale:
            custom_name = get_custom_property(custom_mapping, std, "name")
            if custom_name:
                return custom_name

    try:
        std_locale = standardize_locale(locale)
        parsed = Locale.parse(std_locale, sep="-")

        # Build underscore tag for CLDR compound name lookup
        underscore_tag = parsed.language
        if parsed.script:
            underscore_tag += f"_{parsed.script}"
        if parsed.territory:
            underscore_tag += f"_{parsed.territory}"

        # Try lookup in display locale's language table for compound names
        try:
            display_locale = Locale.parse(default_locale, sep="-")
            # Compound names like "Austrian German", "American English"
            compound = display_locale.languages.get(underscore_tag)
            if compound:
                return compound
        except Exception:
            pass

        # Fall back to get_display_name
        name = parsed.get_display_name(default_locale)
        if name:
            return name

        return ""
    except Exception:
        return ""



"""List formatting using Babel.

Ports ``_formatList`` and ``_formatListToParts`` from the JS core library.
"""

from __future__ import annotations

from typing import Any, TypeVar

from babel.lists import format_list as babel_format_list

from generaltranslation.formatting._helpers import _resolve_babel_locale

T = TypeVar("T")

# Map JS ListFormat {type, style} to Babel style parameter.
# Babel 2.18+ supports: "standard", "standard-short", "standard-narrow",
# "or", "or-short", "or-narrow", "unit", "unit-short", "unit-narrow"
_STYLE_MAP: dict[tuple[str, str], str] = {
    ("conjunction", "long"): "standard",
    ("conjunction", "short"): "standard-short",
    ("conjunction", "narrow"): "standard-narrow",
    ("disjunction", "long"): "or",
    ("disjunction", "short"): "or-short",
    ("disjunction", "narrow"): "or-narrow",
    ("unit", "long"): "unit",
    ("unit", "short"): "unit-short",
    ("unit", "narrow"): "unit-narrow",
}


def format_list(
    value: list[Any],
    locales: str | list[str] | None = None,
    options: dict | None = None,
) -> str:
    """Format a list of items according to the given locale and options.

    Args:
        value: The list of items to format. Items are converted to strings.
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Formatting options (snake_case).

            - ``type``: ``"conjunction"`` (default), ``"disjunction"``, ``"unit"``.
            - ``style``: ``"long"`` (default), ``"short"``, ``"narrow"``.

    Returns:
        The formatted list string.
    """
    if options is None:
        options = {}

    locale = _resolve_babel_locale(locales)
    list_type = options.get("type", "conjunction")
    list_style = options.get("style", "long")

    babel_style = _STYLE_MAP.get((list_type, list_style), "standard")

    str_items = [str(item) for item in value]
    return babel_format_list(str_items, style=babel_style, locale=locale)  # type: ignore[arg-type]


def format_list_to_parts(
    value: list[T],
    locales: str | list[str] | None = None,
    options: dict | None = None,
) -> list[T | str]:
    """Format a list and return interleaved parts with separators.

    Uses placeholder-based splitting to extract separators from Babel's
    formatted output, then interleaves them with the original typed values.

    This mirrors the JS ``_formatListToParts`` which uses
    ``Intl.ListFormat.formatToParts()``.

    Args:
        value: The list of items to format.
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Formatting options (same as :func:`format_list`).

    Returns:
        A list alternating between original values and separator strings.
    """
    if not value:
        return []
    if len(value) == 1:
        return list(value)

    if options is None:
        options = {}

    locale = _resolve_babel_locale(locales)
    list_type = options.get("type", "conjunction")
    list_style = options.get("style", "long")
    babel_style = _STYLE_MAP.get((list_type, list_style), "standard")

    # Use unique placeholders to identify element positions
    placeholder = "\x00"
    placeholders = [f"{placeholder}{i}{placeholder}" for i in range(len(value))]
    formatted = babel_format_list(placeholders, style=babel_style, locale=locale)  # type: ignore[arg-type]

    # Split by placeholders to extract separators
    parts: list[T | str] = []
    remaining = formatted
    for i, val in enumerate(value):
        ph = f"{placeholder}{i}{placeholder}"
        idx = remaining.find(ph)
        if idx > 0:
            # There's a separator before this element
            parts.append(remaining[:idx])
        parts.append(val)
        remaining = remaining[idx + len(ph) :]

    # Any trailing text after the last placeholder
    if remaining:
        parts.append(remaining)

    return parts



"""Query branch data endpoint."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._request import api_request


async def query_branch_data(
    query: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Query branch information from the API."""
    return await api_request(
        config,
        "/v2/project/branches/info",
        body=query,
    )



"""Process file moves endpoint."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._batch import process_batches
from generaltranslation.translate._request import api_request


async def process_file_moves(
    moves: list[dict[str, Any]],
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Process file moves by cloning source files and translations."""
    if not moves:
        return {
            "results": [],
            "summary": {"total": 0, "succeeded": 0, "failed": 0},
        }

    async def _process_batch(batch: list[dict[str, Any]]) -> list[Any]:
        result = await api_request(
            config,
            "/v2/project/files/moves",
            body={
                "branchId": options.get("branch_id", options.get("branchId")),
                "moves": batch,
            },
            timeout=options.get("timeout"),
        )
        return result.get("results", [])

    batch_result = await process_batches(moves, _process_batch, batch_size=100)

    succeeded = sum(1 for r in batch_result["data"] if r.get("success"))
    failed = sum(1 for r in batch_result["data"] if not r.get("success"))

    return {
        "results": batch_result["data"],
        "summary": {
            "total": len(moves),
            "succeeded": succeeded,
            "failed": failed,
        },
    }



"""Shared types for the locales subpackage.

Ports the locale-related type definitions from the JS core library:
- ``LocaleProperties`` dataclass (from ``getLocaleProperties.ts``)
- ``CustomMapping`` / ``CustomRegionMapping`` type aliases
- ``PluralType`` literal and ``PLURAL_FORMS`` constant (from ``settings/plurals.ts``)
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

# ---------------------------------------------------------------------------
# Plural forms
# ---------------------------------------------------------------------------

PluralType = Literal[
    "singular",
    "plural",
    "dual",
    "zero",
    "one",
    "two",
    "few",
    "many",
    "other",
]
"""All recognised plural-form branch names."""

PLURAL_FORMS: list[PluralType] = [
    "singular",
    "plural",
    "dual",
    "zero",
    "one",
    "two",
    "few",
    "many",
    "other",
]
"""Ordered list of accepted plural forms, matching the JS ``pluralForms`` constant."""


def is_accepted_plural_form(form: str) -> bool:
    """Return ``True`` if *form* is a recognised plural-form name."""
    return form in PLURAL_FORMS


# ---------------------------------------------------------------------------
# Locale properties
# ---------------------------------------------------------------------------


@dataclass
class LocaleProperties:
    """Comprehensive metadata about a BCP 47 locale.

    This mirrors the JS ``LocaleProperties`` type from ``getLocaleProperties.ts``.
    All string fields default to ``""`` so callers can rely on safe access.

    Example for ``de-AT`` with *default_locale* ``"en-US"``:

    =========================================  =====================================
    Field                                      Example value
    =========================================  =====================================
    ``code``                                   ``"de-AT"``
    ``name``                                   ``"Austrian German"``
    ``native_name``                            ``"Österreichisches Deutsch"``
    ``language_code``                          ``"de"``
    ``language_name``                          ``"German"``
    ``native_language_name``                   ``"Deutsch"``
    ``name_with_region_code``                  ``"German (AT)"``
    ``native_name_with_region_code``           ``"Deutsch (AT)"``
    ``region_code``                            ``"AT"``
    ``region_name``                            ``"Austria"``
    ``native_region_name``                     ``"Österreich"``
    ``script_code``                            ``"Latn"``
    ``script_name``                            ``"Latin"``
    ``native_script_name``                     ``"Lateinisch"``
    ``maximized_code``                         ``"de-Latn-AT"``
    ``maximized_name``                         ``"Austrian German (Latin)"``
    ``native_maximized_name``                  ``"Österr. Deutsch (Lateinisch)"``
    ``minimized_code``                         ``"de-AT"``
    ``minimized_name``                         ``"Austrian German"``
    ``native_minimized_name``                  ``"Österreichisches Deutsch"``
    ``emoji``                                  ``"🇦🇹"``
    =========================================  =====================================
    """

    code: str = ""
    name: str = ""
    native_name: str = ""

    language_code: str = ""
    language_name: str = ""
    native_language_name: str = ""

    name_with_region_code: str = ""
    native_name_with_region_code: str = ""

    region_code: str = ""
    region_name: str = ""
    native_region_name: str = ""

    script_code: str = ""
    script_name: str = ""
    native_script_name: str = ""

    maximized_code: str = ""
    maximized_name: str = ""
    native_maximized_name: str = ""

    minimized_code: str = ""
    minimized_name: str = ""
    native_minimized_name: str = ""

    emoji: str = ""


# ---------------------------------------------------------------------------
# Custom mapping types
# ---------------------------------------------------------------------------

CustomMapping = dict[str, str | dict[str, str]]
"""Maps locale aliases to either a display name (``str``) or a partial
``LocaleProperties``-like dict.  When the dict contains a ``"code"`` key the
alias is treated as a redirect to that canonical locale.

Examples::

    # String form — just a custom name
    {"es-LAM": "Latin American Spanish"}

    # Dict form — partial properties
    {"en-simple": {"name": "Simple English", "emoji": "🎯"}}

    # Canonical redirect — alias resolves to real locale
    {"pt-BR-custom": {"code": "pt-BR", "name": "Custom Brazilian Portuguese"}}
"""

CustomRegionMapping = dict[str, dict[str, str]]
"""Maps region codes to custom ``{"name": ..., "emoji": ..., "locale": ...}``
overrides used by :func:`get_region_properties`."""



"""Generate request headers for API calls."""

from __future__ import annotations

from typing import Any

from generaltranslation._settings import API_VERSION


def generate_request_headers(config: dict[str, Any], *, exclude_content_type: bool = False) -> dict[str, str]:
    headers: dict[str, str] = {}
    if not exclude_content_type:
        headers["Content-Type"] = "application/json"
    headers["x-gt-project-id"] = config.get("project_id", "")

    api_key = config.get("api_key", "")
    if api_key:
        if api_key.startswith("gtx-internal-"):
            headers["x-gt-internal-api-key"] = api_key
        else:
            headers["x-gt-api-key"] = api_key

    headers["gt-api-version"] = API_VERSION
    return headers



"""Minimize BCP 47 locale tags using CLDR likely-subtags.

Babel provides no built-in ``minimize()`` (JS has
``Intl.Locale.minimize()``).  This module implements it using the CLDR
``likely_subtags`` table shipped with Babel.

Algorithm
---------
Find the shortest tag (language, language-script, or
language-territory) whose maximized form equals the fully maximized
input.  This is the inverse of maximize.

Examples
--------
>>> minimize_locale("en-US")
'en'
>>> minimize_locale("en-GB")
'en-GB'
>>> minimize_locale("zh-Hans-CN")
'zh'
>>> minimize_locale("zh-Hant-TW")
'zh-Hant'
"""

from __future__ import annotations

import logging
from typing import cast

from babel import Locale
from babel.core import get_global

logger = logging.getLogger(__name__)

# CLDR likely-subtags table: maps partial locale tags to their most
# likely fully-qualified form.
# e.g. "en" → "en_Latn_US", "zh_Hant" → "zh_Hant_TW"
_likely_subtags: dict[str, str] = cast(dict[str, str], get_global("likely_subtags"))


def _parse_locale(
    locale: str,
) -> tuple[str, str | None, str | None]:
    """Parse a BCP 47 tag into (language, script, territory).

    Accepts both hyphen-separated (BCP 47) and underscore-separated
    (CLDR/Babel) formats.
    """
    parsed = Locale.parse(locale, sep="-" if "-" in locale else "_")
    return parsed.language, parsed.script, parsed.territory


def _lookup_likely(
    lang: str,
    script: str | None = None,
    terr: str | None = None,
) -> tuple[str, str | None, str | None]:
    """Find the most likely fully-qualified subtags from CLDR.

    Tries progressively broader keys until a match is found:
    ``lang_script_terr`` -> ``lang_script`` -> ``lang_terr`` -> ``lang``.
    """
    candidates: list[str] = []
    if script and terr:
        candidates.append(f"{lang}_{script}_{terr}")
    if script:
        candidates.append(f"{lang}_{script}")
    if terr:
        candidates.append(f"{lang}_{terr}")
    candidates.append(lang)

    for key in candidates:
        if key in _likely_subtags:
            parts = _likely_subtags[key].split("_")
            return (
                parts[0],
                parts[1] if len(parts) > 1 else None,
                parts[2] if len(parts) > 2 else None,
            )
    return lang, script, terr


def _maximize(
    lang: str,
    script: str | None,
    terr: str | None,
) -> tuple[str, str | None, str | None]:
    """Maximize parsed components, preserving original values."""
    _, likely_script, likely_terr = _lookup_likely(lang, script, terr)
    return (lang, script or likely_script, terr or likely_terr)


def minimize_locale(locale: str) -> str:
    """Remove likely subtags to produce the shortest equivalent tag.

    Finds the shortest tag (language-only, language-script, or
    language-territory) whose maximized form equals the fully
    maximized input.

    This mirrors ``Intl.Locale.prototype.minimize()`` in JavaScript.

    Args:
        locale: A BCP 47 locale tag (e.g. ``"en-US"``,
            ``"zh-Hans-CN"``).

    Returns:
        The minimized tag with hyphens (e.g. ``"en"``, ``"zh"``).
        Returns the original string on parse failure.

    Examples:
        >>> minimize_locale("en-US")
        'en'
        >>> minimize_locale("en-GB")
        'en-GB'
        >>> minimize_locale("de-AT")
        'de-AT'
        >>> minimize_locale("zh-Hans-CN")
        'zh'
        >>> minimize_locale("zh-Hant-TW")
        'zh-Hant'
    """
    try:
        lang, script, terr = _parse_locale(locale)
        full = _maximize(lang, script, terr)

        # Try language only
        if _maximize(lang, None, None) == full:
            return lang

        # Try language + script
        if script and _maximize(lang, script, None) == full:
            return f"{lang}-{script}"

        # Try language + territory
        if terr and _maximize(lang, None, terr) == full:
            return f"{lang}-{terr}"

        # Cannot reduce — return all original parts
        parts = [lang]
        if script:
            parts.append(script)
        if terr:
            parts.append(terr)
        return "-".join(parts)
    except Exception:
        logger.debug("Failed to minimize locale '%s'", locale)
        return locale



"""Internal helpers for custom locale mapping lookups.

Ports ``customLocaleMapping.ts`` from the JS core library.
Custom mappings allow users to define locale aliases (e.g. ``"es-LAM"``
→ ``"es-419"``) and override display properties without modifying
standard BCP 47 codes.
"""

from __future__ import annotations

from generaltranslation.locales._types import CustomMapping


def get_custom_property(
    custom_mapping: CustomMapping,
    locale: str,
    property_name: str,
) -> str | None:
    """Look up a single property from a custom mapping entry.

    If the mapping entry for *locale* is a plain string, only ``"name"``
    is available.  If it is a dict, the requested *property_name* key is
    returned when present.

    Args:
        custom_mapping: The custom locale mapping dict.
        locale: The locale key to look up in the mapping.
        property_name: The property to extract (e.g. ``"name"``,
            ``"emoji"``, ``"code"``).

    Returns:
        The property value if found, otherwise ``None``.

    Examples:
        >>> mapping = {"es-LAM": "Latin American Spanish"}
        >>> get_custom_property(mapping, "es-LAM", "name")
        'Latin American Spanish'
        >>> get_custom_property(mapping, "es-LAM", "emoji") is None
        True
    """
    entry = custom_mapping.get(locale)
    if entry is None:
        return None
    if isinstance(entry, str):
        return entry if property_name == "name" else None
    if isinstance(entry, dict):
        return entry.get(property_name)
    return None


def should_use_canonical_locale(
    locale: str,
    custom_mapping: CustomMapping,
) -> bool:
    """Check whether *locale* should be resolved to a canonical locale via the mapping.

    A custom mapping entry triggers canonical resolution when it is a dict
    containing a ``"code"`` key whose value is itself a valid BCP 47 locale.

    Args:
        locale: The locale alias to check.
        custom_mapping: The custom locale mapping dict.

    Returns:
        ``True`` if the mapping redirects *locale* to a canonical code.

    Examples:
        >>> mapping = {"pt-custom": {"code": "pt-BR", "name": "Custom PT"}}
        >>> should_use_canonical_locale("pt-custom", mapping)
        True
        >>> should_use_canonical_locale("en", mapping)
        False
    """
    entry = custom_mapping.get(locale)
    if isinstance(entry, dict) and "code" in entry:
        return True
    return False



"""Core Python language toolkit for General Translation."""

from generaltranslation._gt import GT
from generaltranslation._id import hash_source, hash_string, hash_template
from generaltranslation._settings import (
    API_VERSION,
    DEFAULT_BASE_URL,
    DEFAULT_CACHE_URL,
    DEFAULT_RUNTIME_API_URL,
    DEFAULT_TIMEOUT,
    LIBRARY_DEFAULT_LOCALE,
)
from generaltranslation.errors import ApiError
from generaltranslation.formatting import (
    CutoffFormat,
    format_currency,
    format_cutoff,
    format_date_time,
    format_list,
    format_list_to_parts,
    format_message,
    format_num,
    format_relative_time,
)
from generaltranslation.locales import (
    PLURAL_FORMS,
    CustomMapping,
    CustomRegionMapping,
    LocaleProperties,
    PluralType,
    determine_locale,
    get_locale_direction,
    get_locale_emoji,
    get_locale_name,
    get_locale_properties,
    get_plural_form,
    get_region_properties,
    is_accepted_plural_form,
    is_same_dialect,
    is_same_language,
    is_superset_locale,
    is_valid_locale,
    requires_translation,
    resolve_alias_locale,
    resolve_canonical_locale,
    standardize_locale,
)
from generaltranslation.static import (
    VAR_IDENTIFIER,
    VAR_NAME_IDENTIFIER,
    condense_vars,
    declare_static,
    declare_var,
    decode_vars,
    extract_vars,
    index_vars,
    sanitize_var,
)

__all__ = [
    # Core class
    "GT",
    # Constants
    "LIBRARY_DEFAULT_LOCALE",
    "DEFAULT_BASE_URL",
    "DEFAULT_CACHE_URL",
    "DEFAULT_RUNTIME_API_URL",
    "DEFAULT_TIMEOUT",
    "API_VERSION",
    "PLURAL_FORMS",
    # Errors
    "ApiError",
    # ID / Hashing
    "hash_string",
    "hash_source",
    "hash_template",
    # Formatting
    "format_num",
    "format_currency",
    "format_date_time",
    "format_list",
    "format_list_to_parts",
    "format_relative_time",
    "format_message",
    "format_cutoff",
    "CutoffFormat",
    # Locale types
    "LocaleProperties",
    "CustomMapping",
    "CustomRegionMapping",
    "PluralType",
    "is_accepted_plural_form",
    # Locale functions
    "is_valid_locale",
    "standardize_locale",
    "get_locale_properties",
    "get_locale_name",
    "get_locale_emoji",
    "get_locale_direction",
    "get_region_properties",
    "determine_locale",
    "resolve_canonical_locale",
    "resolve_alias_locale",
    "is_same_language",
    "is_same_dialect",
    "is_superset_locale",
    "requires_translation",
    "get_plural_form",
    # Static
    "VAR_IDENTIFIER",
    "VAR_NAME_IDENTIFIER",
    "sanitize_var",
    "declare_var",
    "declare_static",
    "decode_vars",
    "extract_vars",
    "index_vars",
    "condense_vars",
]



"""Get orphaned files endpoint."""

from __future__ import annotations

import asyncio
from typing import Any

from generaltranslation.translate._batch import create_batches
from generaltranslation.translate._request import api_request


async def get_orphaned_files(
    branch_id: str,
    file_ids: list[str],
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Get orphaned files for a branch.

    Files that exist on the branch but whose fileIds are not in the provided list.
    """

    async def _make_request(batch_file_ids: list[str]) -> dict[str, Any]:
        return await api_request(
            config,
            "/v2/project/files/orphaned",
            body={"branchId": branch_id, "fileIds": batch_file_ids},
            timeout=options.get("timeout"),
        )

    if not file_ids:
        return await _make_request([])

    batches = create_batches(file_ids, 100)
    batch_results = await asyncio.gather(*(_make_request(batch) for batch in batches))

    if len(batch_results) == 1:
        return batch_results[0]

    # Find intersection of orphaned files across all batches
    orphaned_map: dict[str, dict[str, Any]] = {}
    for orphan in batch_results[0].get("orphanedFiles", []):
        orphaned_map[orphan["fileId"]] = orphan

    for i in range(1, len(batch_results)):
        batch_orphan_ids = {f["fileId"] for f in batch_results[i].get("orphanedFiles", [])}
        for file_id in list(orphaned_map.keys()):
            if file_id not in batch_orphan_ids:
                del orphaned_map[file_id]

    return {"orphanedFiles": list(orphaned_map.values())}



"""Main GT class."""

from __future__ import annotations

import os
from typing import Any

from generaltranslation._settings import LIBRARY_DEFAULT_LOCALE
from generaltranslation.errors import (
    invalid_locale_error,
    invalid_locales_error,
    no_api_key_error,
    no_project_id_error,
    no_source_locale_error,
    no_target_locale_error,
)
from generaltranslation.formatting import (
    format_currency,
    format_cutoff,
    format_date_time,
    format_list,
    format_list_to_parts,
    format_message,
    format_num,
    format_relative_time,
)
from generaltranslation.locales import (
    CustomMapping,
    CustomRegionMapping,
    determine_locale,
    get_locale_direction,
    get_locale_emoji,
    get_locale_name,
    get_locale_properties,
    get_region_properties,
    is_same_dialect,
    is_same_language,
    is_superset_locale,
    is_valid_locale,
    requires_translation,
    resolve_alias_locale,
    resolve_canonical_locale,
    standardize_locale,
)
from generaltranslation.translate import (
    check_job_status as _check_job_status,
)
from generaltranslation.translate import (
    create_branch as _create_branch,
)
from generaltranslation.translate import (
    download_file_batch as _download_file_batch,
)
from generaltranslation.translate import (
    enqueue_files as _enqueue_files,
)
from generaltranslation.translate import (
    get_orphaned_files as _get_orphaned_files,
)
from generaltranslation.translate import (
    get_project_data as _get_project_data,
)
from generaltranslation.translate import (
    process_file_moves as _process_file_moves,
)
from generaltranslation.translate import (
    query_branch_data as _query_branch_data,
)
from generaltranslation.translate import (
    query_file_data as _query_file_data,
)
from generaltranslation.translate import (
    query_source_file as _query_source_file,
)
from generaltranslation.translate import (
    setup_project as _setup_project,
)
from generaltranslation.translate import (
    submit_user_edit_diffs as _submit_user_edit_diffs,
)
from generaltranslation.translate import (
    translate_many as _translate_many,
)
from generaltranslation.translate import (
    upload_source_files as _upload_source_files,
)
from generaltranslation.translate import (
    upload_translations as _upload_translations,
)


class GT:
    """General Translation core driver class.

    A comprehensive toolkit for handling internationalisation and localisation.
    """

    def __init__(
        self,
        *,
        api_key: str | None = None,
        dev_api_key: str | None = None,
        project_id: str | None = None,
        base_url: str | None = None,
        source_locale: str | None = None,
        target_locale: str | None = None,
        locales: list[str] | None = None,
        custom_mapping: CustomMapping | None = None,
    ) -> None:
        # Read environment variables first
        self.api_key: str | None = api_key or os.environ.get("GT_API_KEY") or None
        self.dev_api_key: str | None = dev_api_key or os.environ.get("GT_DEV_API_KEY") or None
        self.project_id: str | None = project_id or os.environ.get("GT_PROJECT_ID") or None

        self.base_url: str | None = None
        self.source_locale: str | None = None
        self.target_locale: str | None = None
        self.locales: list[str] | None = None
        self.custom_mapping: CustomMapping | None = None
        self.reverse_custom_mapping: dict[str, str] | None = None
        self.custom_region_mapping: CustomRegionMapping | None = None
        self._rendering_locales: list[str] = []

        self.set_config(
            api_key=api_key,
            dev_api_key=dev_api_key,
            project_id=project_id,
            base_url=base_url,
            source_locale=source_locale,
            target_locale=target_locale,
            locales=locales,
            custom_mapping=custom_mapping,
        )

    def set_config(
        self,
        *,
        api_key: str | None = None,
        dev_api_key: str | None = None,
        project_id: str | None = None,
        base_url: str | None = None,
        source_locale: str | None = None,
        target_locale: str | None = None,
        locales: list[str] | None = None,
        custom_mapping: CustomMapping | None = None,
    ) -> None:
        """Update instance configuration."""
        if api_key:
            self.api_key = api_key
        if dev_api_key:
            self.dev_api_key = dev_api_key
        if project_id:
            self.project_id = project_id

        # Standardise locales
        if source_locale:
            self.source_locale = standardize_locale(source_locale)
            if not is_valid_locale(self.source_locale, custom_mapping):
                raise ValueError(invalid_locale_error(self.source_locale))

        if target_locale:
            self.target_locale = standardize_locale(target_locale)
            if not is_valid_locale(self.target_locale, custom_mapping):
                raise ValueError(invalid_locale_error(self.target_locale))

        # Rendering locales
        self._rendering_locales = []
        if self.source_locale:
            self._rendering_locales.append(self.source_locale)
        if self.target_locale:
            self._rendering_locales.append(self.target_locale)
        self._rendering_locales.append(LIBRARY_DEFAULT_LOCALE)

        # Validate locales list
        if locales is not None:
            result: list[str] = []
            invalid: list[str] = []
            for loc in locales:
                std = standardize_locale(loc)
                if is_valid_locale(std):
                    result.append(std)
                else:
                    invalid.append(loc)
            if invalid:
                raise ValueError(invalid_locales_error(invalid))
            self.locales = result

        if base_url:
            self.base_url = base_url
        if custom_mapping:
            self.custom_mapping = custom_mapping
            self.reverse_custom_mapping = {
                v["code"]: k for k, v in custom_mapping.items() if v and isinstance(v, dict) and "code" in v
            }

    # -------------- Private Methods -------------- #

    def _get_translation_config(self) -> dict[str, Any]:
        return {
            "base_url": self.base_url,
            "api_key": self.api_key or self.dev_api_key or "",
            "project_id": self.project_id or "",
        }

    def _validate_auth(self, fn_name: str) -> None:
        errors: list[str] = []
        if not self.api_key and not self.dev_api_key:
            errors.append(no_api_key_error(fn_name))
        if not self.project_id:
            errors.append(no_project_id_error(fn_name))
        if errors:
            raise ValueError("\n".join(errors))

    # -------------- Branch Methods -------------- #

    async def query_branch_data(self, query: dict[str, Any]) -> dict[str, Any]:
        """Query branch information from the API."""
        self._validate_auth("query_branch_data")
        return await _query_branch_data(query, self._get_translation_config())

    async def create_branch(self, query: dict[str, Any]) -> dict[str, Any]:
        """Create a new branch in the API."""
        self._validate_auth("create_branch")
        return await _create_branch(query, self._get_translation_config())

    async def process_file_moves(
        self,
        moves: list[dict[str, Any]],
        options: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Process file moves by cloning source files and translations."""
        self._validate_auth("process_file_moves")
        return await _process_file_moves(moves, options or {}, self._get_translation_config())

    async def get_orphaned_files(
        self,
        branch_id: str,
        file_ids: list[str],
        options: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get orphaned files for a branch."""
        self._validate_auth("get_orphaned_files")
        return await _get_orphaned_files(branch_id, file_ids, options or {}, self._get_translation_config())

    # -------------- Translation Methods -------------- #

    async def setup_project(
        self,
        files: list[dict[str, Any]],
        options: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Enqueue files for project setup."""
        self._validate_auth("setup_project")
        opts = dict(options) if options else {}
        if opts.get("locales"):
            opts["locales"] = [self.resolve_canonical_locale(loc) for loc in opts["locales"]]
        return await _setup_project(files, self._get_translation_config(), opts)

    async def check_job_status(
        self,
        job_ids: list[str],
        timeout_ms: int | None = None,
    ) -> list[dict[str, Any]]:
        """Check job statuses."""
        self._validate_auth("check_job_status")
        return await _check_job_status(job_ids, self._get_translation_config(), timeout_ms)

    async def enqueue_files(
        self,
        files: list[dict[str, Any]],
        options: dict[str, Any],
    ) -> dict[str, Any]:
        """Enqueue translation jobs for previously uploaded source files."""
        self._validate_auth("enqueue_files")
        merged = dict(options)
        merged.setdefault("source_locale", self.source_locale)
        merged.setdefault("sourceLocale", self.source_locale)
        if not merged.get("target_locales") and not merged.get("targetLocales"):
            if self.target_locale:
                merged["target_locales"] = [self.target_locale]

        src = merged.get("source_locale") or merged.get("sourceLocale")
        if not src:
            raise ValueError(no_source_locale_error("enqueue_files"))
        tgts = merged.get("target_locales") or merged.get("targetLocales") or []
        if not tgts:
            raise ValueError(no_target_locale_error("enqueue_files"))

        # Resolve canonical locales
        resolved_targets = [self.resolve_canonical_locale(loc) for loc in tgts]
        merged["target_locales"] = resolved_targets
        merged["targetLocales"] = resolved_targets

        return await _enqueue_files(files, merged, self._get_translation_config())

    async def submit_user_edit_diffs(self, payload: dict[str, Any]) -> None:
        """Submit user edit diffs."""
        self._validate_auth("submit_user_edit_diffs")
        normalized = dict(payload)
        if normalized.get("diffs"):
            normalized["diffs"] = [
                {**d, "locale": self.resolve_canonical_locale(d.get("locale"))} for d in normalized["diffs"]
            ]
        await _submit_user_edit_diffs(normalized, self._get_translation_config())

    async def query_file_data(
        self,
        data: dict[str, Any],
        options: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Query data about one or more source or translation files."""
        self._validate_auth("query_file_data")
        data = dict(data)
        tf_key = "translated_files" if "translated_files" in data else "translatedFiles"
        if data.get(tf_key):
            data[tf_key] = [
                {**item, "locale": self.resolve_canonical_locale(item.get("locale"))} for item in data[tf_key]
            ]

        result = await _query_file_data(data, options or {}, self._get_translation_config())

        # Resolve alias locales in response
        if result.get("translatedFiles"):
            result["translatedFiles"] = [
                {
                    **item,
                    **({"locale": self.resolve_alias_locale(item["locale"])} if item.get("locale") else {}),
                }
                for item in result["translatedFiles"]
            ]
        if result.get("sourceFiles"):
            result["sourceFiles"] = [
                {
                    **item,
                    **(
                        {"sourceLocale": self.resolve_alias_locale(item["sourceLocale"])}
                        if item.get("sourceLocale")
                        else {}
                    ),
                    "locales": [self.resolve_alias_locale(loc) for loc in item.get("locales", [])],
                }
                for item in result["sourceFiles"]
            ]
        return result

    async def query_source_file(
        self,
        data: dict[str, Any],
        options: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get source file and translation information."""
        self._validate_auth("query_source_file")
        result = await _query_source_file(data, options or {}, self._get_translation_config())
        if result.get("translations"):
            result["translations"] = [
                {
                    **item,
                    **({"locale": self.resolve_alias_locale(item["locale"])} if item.get("locale") else {}),
                }
                for item in result["translations"]
            ]
        sf = result.get("sourceFile", {})
        if sf.get("locales"):
            sf["locales"] = [self.resolve_alias_locale(loc) for loc in sf["locales"]]
        if sf.get("sourceLocale"):
            sf["sourceLocale"] = self.resolve_alias_locale(sf["sourceLocale"])
        return result

    async def get_project_data(
        self,
        project_id: str,
        options: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get project data for a given project ID."""
        self._validate_auth("get_project_data")
        result = await _get_project_data(project_id, options or {}, self._get_translation_config())
        if result.get("currentLocales"):
            result["currentLocales"] = [self.resolve_alias_locale(loc) for loc in result["currentLocales"]]
        if result.get("defaultLocale"):
            result["defaultLocale"] = self.resolve_alias_locale(result["defaultLocale"])
        return result

    async def download_file(
        self,
        file: dict[str, Any],
        options: dict[str, Any] | None = None,
    ) -> str:
        """Download a single file."""
        self._validate_auth("download_file")
        request = {
            "fileId": file.get("file_id", file.get("fileId", "")),
            "branchId": file.get("branch_id", file.get("branchId")),
            "locale": self.resolve_canonical_locale(file.get("locale")),
            "versionId": file.get("version_id", file.get("versionId")),
        }
        result = await _download_file_batch([request], options or {}, self._get_translation_config())
        return result["data"][0]["data"]

    async def download_file_batch(
        self,
        requests: list[dict[str, Any]],
        options: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Download multiple files in a batch."""
        self._validate_auth("download_file_batch")
        mapped_requests = [{**r, "locale": self.resolve_canonical_locale(r.get("locale"))} for r in requests]
        result = await _download_file_batch(mapped_requests, options or {}, self._get_translation_config())
        files = [
            {
                **f,
                **({"locale": self.resolve_alias_locale(f["locale"])} if f.get("locale") else {}),
            }
            for f in result["data"]
        ]
        return {"files": files, "count": result["count"]}

    async def translate(
        self,
        source: Any,
        options: str | dict[str, Any],
        timeout: int | None = None,
    ) -> dict[str, Any]:
        """Translate a single source string."""
        if isinstance(options, str):
            options = {"target_locale": options}
        self._validate_auth("translate")

        target_locale = options.get("target_locale") or options.get("targetLocale") or self.target_locale
        if not target_locale:
            raise ValueError(no_target_locale_error("translate"))
        target_locale = self.resolve_canonical_locale(target_locale)

        source_locale = self.resolve_canonical_locale(
            options.get("source_locale") or options.get("sourceLocale") or self.source_locale or LIBRARY_DEFAULT_LOCALE
        )

        results = await _translate_many(
            [source],
            {**options, "target_locale": target_locale, "source_locale": source_locale},
            self._get_translation_config(),
            timeout,
        )
        assert isinstance(results, list)
        return results[0]

    async def translate_many(
        self,
        sources: list[Any] | dict[str, Any],
        options: str | dict[str, Any],
        timeout: int | None = None,
    ) -> list[dict[str, Any]] | dict[str, dict[str, Any]]:
        """Translate multiple source strings."""
        if isinstance(options, str):
            options = {"target_locale": options}
        self._validate_auth("translate_many")

        target_locale = options.get("target_locale") or options.get("targetLocale") or self.target_locale
        if not target_locale:
            raise ValueError(no_target_locale_error("translate_many"))
        target_locale = self.resolve_canonical_locale(target_locale)

        source_locale = self.resolve_canonical_locale(
            options.get("source_locale") or options.get("sourceLocale") or self.source_locale or LIBRARY_DEFAULT_LOCALE
        )

        return await _translate_many(
            sources,
            {**options, "target_locale": target_locale, "source_locale": source_locale},
            self._get_translation_config(),
            timeout,
        )

    async def upload_source_files(
        self,
        files: list[dict[str, Any]],
        options: dict[str, Any],
    ) -> dict[str, Any]:
        """Upload source files to the translation service."""
        self._validate_auth("upload_source_files")
        merged = dict(options)
        merged["source_locale"] = self.resolve_canonical_locale(
            options.get("source_locale") or options.get("sourceLocale") or self.source_locale or LIBRARY_DEFAULT_LOCALE
        )
        mapped_files = [
            {
                **f,
                "source": {
                    **f["source"],
                    "locale": self.resolve_canonical_locale(f["source"].get("locale")),
                },
            }
            for f in files
        ]
        result = await _upload_source_files(mapped_files, merged, self._get_translation_config())
        return {
            "uploadedFiles": result["data"],
            "count": result["count"],
            "message": f"Successfully uploaded {result['count']} files in {result['batch_count']} batch(es)",
        }

    async def upload_translations(
        self,
        files: list[dict[str, Any]],
        options: dict[str, Any],
    ) -> dict[str, Any]:
        """Upload translation files."""
        self._validate_auth("upload_translations")
        merged = dict(options)
        merged.setdefault("source_locale", self.source_locale)
        if not merged.get("source_locale"):
            raise ValueError(no_source_locale_error("upload_translations"))

        mapped_files = [
            {
                **f,
                "translations": [
                    {**t, "locale": self.resolve_canonical_locale(t.get("locale"))} for t in f.get("translations", [])
                ],
            }
            for f in files
        ]
        result = await _upload_translations(mapped_files, merged, self._get_translation_config())
        return {
            "uploadedFiles": result["data"],
            "count": result["count"],
            "message": f"Successfully uploaded {result['count']} files in {result['batch_count']} batch(es)",
        }

    # -------------- Formatting -------------- #

    def format_cutoff(
        self,
        value: str,
        *,
        locales: str | list[str] | None = None,
        options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> str:
        """Format a string with cutoff behaviour."""
        opts = dict(options or {})
        opts.update(kwargs)
        return format_cutoff(value, locales=locales or self._rendering_locales, options=opts)

    def format_message(
        self,
        message: str,
        *,
        locales: str | list[str] | None = None,
        variables: dict[str, Any] | None = None,
    ) -> str:
        """Format a message with variables."""
        return format_message(
            message,
            locales=locales or self._rendering_locales,
            variables=variables,
        )

    def format_num(self, number: float | int, **kwargs: Any) -> str:
        """Format a number."""
        kwargs.setdefault("locales", self._rendering_locales)
        return format_num(number, **kwargs)

    def format_date_time(self, date: Any, **kwargs: Any) -> str:
        """Format a datetime."""
        kwargs.setdefault("locales", self._rendering_locales)
        return format_date_time(date, **kwargs)

    def format_currency(self, value: float | int, currency: str, **kwargs: Any) -> str:
        """Format a currency value."""
        kwargs.setdefault("locales", self._rendering_locales)
        return format_currency(value, currency, **kwargs)

    def format_list(self, array: list[Any], **kwargs: Any) -> str:
        """Format a list of items."""
        kwargs.setdefault("locales", self._rendering_locales)
        return format_list(array, **kwargs)

    def format_list_to_parts(self, array: list[Any], **kwargs: Any) -> list[Any]:
        """Format a list of items to parts."""
        kwargs.setdefault("locales", self._rendering_locales)
        return format_list_to_parts(array, **kwargs)

    def format_relative_time(self, value: float | int, unit: str, **kwargs: Any) -> str:
        """Format a relative time value."""
        kwargs.setdefault("locales", self._rendering_locales)
        return format_relative_time(value, unit, **kwargs)

    # -------------- Locale Properties -------------- #

    def get_locale_name(self, locale: str | None = None) -> str:
        """Get the display name of a locale code."""
        locale = locale or self.target_locale
        if not locale:
            raise ValueError(no_target_locale_error("get_locale_name"))
        return get_locale_name(locale, self.source_locale, self.custom_mapping)

    def get_locale_emoji(self, locale: str | None = None) -> str:
        """Get emoji for a locale."""
        locale = locale or self.target_locale
        if not locale:
            raise ValueError(no_target_locale_error("get_locale_emoji"))
        return get_locale_emoji(locale, self.custom_mapping)

    def get_locale_properties(self, locale: str | None = None) -> Any:
        """Get detailed locale properties."""
        locale = locale or self.target_locale
        if not locale:
            raise ValueError(no_target_locale_error("get_locale_properties"))
        return get_locale_properties(locale, self.source_locale, self.custom_mapping)

    def get_region_properties(
        self,
        region: str | None = None,
        custom_mapping: CustomRegionMapping | None = None,
    ) -> dict[str, str]:
        """Get region properties."""
        if region is None:
            region = self.get_locale_properties().get("region_code", "")
        if custom_mapping is None:
            if self.custom_mapping and not self.custom_region_mapping:
                crm: CustomRegionMapping = {}
                for loc, lp in self.custom_mapping.items():
                    if lp and isinstance(lp, dict) and lp.get("regionCode") and lp["regionCode"] not in crm:
                        entry: dict[str, Any] = {"locale": loc}
                        if lp.get("regionName"):
                            entry["name"] = lp["regionName"]
                        if lp.get("emoji"):
                            entry["emoji"] = lp["emoji"]
                        crm[lp["regionCode"]] = entry
                self.custom_region_mapping = crm
            custom_mapping = self.custom_region_mapping
        return get_region_properties(region, self.target_locale, custom_mapping)

    def requires_translation(
        self,
        source_locale: str | None = None,
        target_locale: str | None = None,
        approved_locales: list[str] | None = None,
        custom_mapping: CustomMapping | None = None,
    ) -> bool:
        """Check if translation is required."""
        source_locale = source_locale or self.source_locale
        target_locale = target_locale or self.target_locale
        if approved_locales is None:
            approved_locales = self.locales
        if custom_mapping is None:
            custom_mapping = self.custom_mapping
        if not source_locale:
            raise ValueError(no_source_locale_error("requires_translation"))
        if not target_locale:
            raise ValueError(no_target_locale_error("requires_translation"))
        return requires_translation(source_locale, target_locale, approved_locales, custom_mapping)

    def determine_locale(
        self,
        locales: str | list[str],
        approved_locales: list[str] | None = None,
        custom_mapping: CustomMapping | None = None,
    ) -> str | None:
        """Determine the best matching locale."""
        if approved_locales is None:
            approved_locales = self.locales or []
        if custom_mapping is None:
            custom_mapping = self.custom_mapping
        return determine_locale(locales, approved_locales, custom_mapping)

    def get_locale_direction(self, locale: str | None = None) -> str:
        """Get text direction for a locale."""
        locale = locale or self.target_locale
        if not locale:
            raise ValueError(no_target_locale_error("get_locale_direction"))
        return get_locale_direction(locale)

    def is_valid_locale(
        self,
        locale: str | None = None,
        custom_mapping: CustomMapping | None = None,
    ) -> bool:
        """Check if a locale code is valid."""
        locale = locale or self.target_locale
        if custom_mapping is None:
            custom_mapping = self.custom_mapping
        if not locale:
            raise ValueError(no_target_locale_error("is_valid_locale"))
        return is_valid_locale(locale, custom_mapping)

    def resolve_canonical_locale(
        self,
        locale: str | None = None,
        custom_mapping: CustomMapping | None = None,
    ) -> str:
        """Resolve the canonical locale for a given locale."""
        locale = locale or self.target_locale
        if custom_mapping is None:
            custom_mapping = self.custom_mapping
        if not locale:
            raise ValueError(no_target_locale_error("resolve_canonical_locale"))
        return resolve_canonical_locale(locale, custom_mapping)

    def resolve_alias_locale(
        self,
        locale: str,
        custom_mapping: CustomMapping | None = None,
    ) -> str:
        """Resolve the alias locale for a given locale."""
        if custom_mapping is None:
            custom_mapping = self.custom_mapping
        if not locale:
            raise ValueError(no_target_locale_error("resolve_alias_locale"))
        return resolve_alias_locale(locale, custom_mapping)

    def standardize_locale(self, locale: str | None = None) -> str:
        """Standardise a BCP 47 locale code."""
        locale = locale or self.target_locale
        if not locale:
            raise ValueError(no_target_locale_error("standardize_locale"))
        return standardize_locale(locale)

    def is_same_dialect(self, *locales: str | list[str]) -> bool:
        """Check if locale codes represent the same dialect."""
        return is_same_dialect(*locales)

    def is_same_language(self, *locales: str | list[str]) -> bool:
        """Check if locale codes represent the same language."""
        return is_same_language(*locales)

    def is_superset_locale(self, super_locale: str, sub_locale: str) -> bool:
        """Check if a locale is a superset of another locale."""
        return is_superset_locale(super_locale, sub_locale)



"""String cutoff formatting with locale-aware terminators.

Ports ``CutoffFormat`` from the JS core library.
"""

from __future__ import annotations

from typing import cast

from generaltranslation.formatting._helpers import _get_language_code

# Default style when maxChars is set
DEFAULT_CUTOFF_FORMAT_STYLE = "ellipsis"

# Terminator configurations per style and language
TERMINATOR_MAP: dict[str, dict[str, dict[str, str | None]]] = {
    "ellipsis": {
        "fr": {
            "terminator": "\u2026",  # …
            "separator": "\u202f",  # narrow no-break space
        },
        "zh": {
            "terminator": "\u2026\u2026",  # ……
            "separator": None,
        },
        "ja": {
            "terminator": "\u2026\u2026",  # ……
            "separator": None,
        },
        "_default": {
            "terminator": "\u2026",  # …
            "separator": None,
        },
    },
    "none": {
        "_default": {
            "terminator": None,
            "separator": None,
        },
    },
}


class CutoffFormat:
    """Format strings with cutoff behavior and locale-aware terminators.

    Args:
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Formatting options.

            - ``max_chars`` (int | None): Maximum character count.
              ``None`` means no cutoff. Negative values truncate from
              the start.
            - ``style``: ``"ellipsis"`` (default) or ``"none"``.
            - ``terminator``: Override the terminator string.
            - ``separator``: Override the separator between value and
              terminator.
    """

    def __init__(
        self,
        locales: str | list[str] | None = None,
        options: dict | None = None,
    ) -> None:
        if options is None:
            options = {}

        max_chars = options.get("max_chars")
        style_input = options.get("style", DEFAULT_CUTOFF_FORMAT_STYLE)

        # Validate style
        if style_input not in TERMINATOR_MAP:
            raise ValueError(
                f"Invalid cutoff format style: {style_input!r}. Must be one of: {', '.join(TERMINATOR_MAP.keys())}"
            )

        # Resolve terminator options
        style: str | None = None
        preset: dict[str, str | None] | None = None
        if max_chars is not None:
            style = style_input
            lang = _get_language_code(locales)
            style_map = TERMINATOR_MAP[style]
            preset = style_map.get(lang, style_map["_default"])

        terminator = options.get("terminator", preset.get("terminator") if preset else None)
        separator: str | None = None
        if terminator is not None:
            separator = options.get("separator", preset.get("separator") if preset else None)

        # Calculate addition length
        self._addition_length = (len(terminator) if terminator else 0) + (len(separator) if separator else 0)

        # If maxChars doesn't have enough space for terminator+separator, drop them
        if max_chars is not None and abs(max_chars) < self._addition_length:
            terminator = None
            separator = None
            self._addition_length = 0

        self._options = {
            "max_chars": max_chars,
            "style": style,
            "terminator": terminator,
            "separator": separator,
        }

    def format(self, value: str) -> str:
        """Format a string with cutoff behavior.

        Args:
            value: The string to format.

        Returns:
            The formatted string with terminator applied if cutoff occurs.
        """
        return "".join(self.format_to_parts(value))

    def format_to_parts(self, value: str) -> list[str]:
        """Format a string and return the individual parts.

        Args:
            value: The string to format.

        Returns:
            A list of string parts:
            - Positive max_chars: ``[sliced_value, separator?, terminator?]``
            - Negative max_chars: ``[terminator?, separator?, sliced_value]``
            - No cutoff: ``[original_value]``
        """
        max_chars = cast("int | None", self._options["max_chars"])
        terminator = cast("str | None", self._options["terminator"])
        separator = cast("str | None", self._options["separator"])

        # Calculate adjusted cutoff
        if max_chars is None or abs(max_chars) >= len(value):
            adjusted_chars = max_chars
        elif max_chars >= 0:
            adjusted_chars = max(0, max_chars - self._addition_length)
        else:
            adjusted_chars = min(0, max_chars + self._addition_length)

        # Slice the value
        if adjusted_chars is not None and adjusted_chars > -1:
            sliced_value = value[:adjusted_chars]
        else:
            sliced_value = value[adjusted_chars:]

        # No cutoff / no terminator → value only
        if (
            max_chars is None
            or adjusted_chars is None
            or adjusted_chars == 0
            or terminator is None
            or len(value) <= abs(max_chars)
        ):
            return [sliced_value]

        # Postpended cutoff (positive max_chars)
        if adjusted_chars > 0:
            parts = [sliced_value]
            if separator is not None:
                parts.append(separator)
            parts.append(terminator)
            return parts

        # Prepended cutoff (negative max_chars)
        parts = [terminator]
        if separator is not None:
            parts.append(separator)
        parts.append(sliced_value)
        return parts

    def resolved_options(self) -> dict:
        """Return the resolved formatting options.

        Returns:
            A dict with ``max_chars``, ``style``, ``terminator``,
            ``separator``.
        """
        return dict(self._options)


def format_cutoff(
    value: str,
    locales: str | list[str] | None = None,
    options: dict | None = None,
) -> str:
    """Format a string with cutoff behavior.

    Convenience function wrapping :class:`CutoffFormat`.

    Args:
        value: The string to format.
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Cutoff formatting options (see :class:`CutoffFormat`).

    Returns:
        The formatted string with terminator applied if cutoff occurs.
    """
    return CutoffFormat(locales, options).format(value)



"""Best-match locale resolution.

Ports ``determineLocale.ts`` from the JS core library.
Given a list of requested locales and a list of approved locales,
finds the best available match.
"""

from __future__ import annotations

from babel import Locale

from generaltranslation.locales._types import CustomMapping


def determine_locale(
    locales: str | list[str],
    approved_locales: list[str],
    custom_mapping: CustomMapping | None = None,
) -> str | None:
    """Find the best matching locale from *approved_locales*.

    Iterates through *locales* in preference order and, for each one,
    searches *approved_locales* for:

    1. An exact match (after standardisation).
    2. A match on ``language-region`` or ``language-script``.
    3. A match on the minimised locale code.
    4. A same-language fallback using the base language code.

    Both input lists are filtered to valid locales and standardised
    before comparison.

    Args:
        locales: One or more BCP 47 locale tags the caller prefers,
            in descending priority.
        approved_locales: The set of locales the application supports.
        custom_mapping: Optional custom mapping passed through to
            validation and property resolution.

    Returns:
        The best matching locale from *approved_locales*, or ``None``
        if no match is found.

    Examples:
        >>> determine_locale("en-US", ["en", "fr", "de"])
        'en'
        >>> determine_locale(["ja", "en"], ["en-US", "fr"])
        'en-US'
        >>> determine_locale("ko", ["en", "fr"]) is None
        True
    """
    from generaltranslation.locales._is_valid_locale import (
        is_valid_locale,
        standardize_locale,
    )
    from generaltranslation.locales.utils import minimize_locale

    # Normalize input to list
    if isinstance(locales, str):
        locale_list = [locales]
    else:
        locale_list = list(locales)

    # Filter and standardize requested locales
    valid_requests: list[tuple[str, Locale]] = []
    for loc in locale_list:
        if is_valid_locale(loc, custom_mapping):
            std = standardize_locale(loc)
            try:
                parsed = Locale.parse(std, sep="-")
                valid_requests.append((std, parsed))
            except Exception:
                pass

    # Filter and standardize approved locales, keeping original values
    valid_approved: list[tuple[str, str, Locale]] = []
    for loc in approved_locales:
        if is_valid_locale(loc, custom_mapping):
            std = standardize_locale(loc)
            try:
                parsed = Locale.parse(std, sep="-")
                valid_approved.append((loc, std, parsed))
            except Exception:
                pass

    if not valid_requests or not valid_approved:
        return None

    for req_std, req_parsed in valid_requests:
        # 1. Exact match
        for orig, app_std, app_parsed in valid_approved:
            if req_std == app_std:
                return orig

        # Find same-language candidates
        same_lang = [
            (orig, app_std, app_parsed)
            for orig, app_std, app_parsed in valid_approved
            if app_parsed.language == req_parsed.language
        ]

        if not same_lang:
            continue

        # 2. Match on language-region or language-script
        if req_parsed.territory:
            for orig, app_std, app_parsed in same_lang:
                if app_parsed.territory == req_parsed.territory:
                    return orig

        if req_parsed.script:
            for orig, app_std, app_parsed in same_lang:
                if app_parsed.script == req_parsed.script:
                    return orig

        # 3. Match on minimized locale
        req_min = minimize_locale(req_std)
        for orig, app_std, app_parsed in same_lang:
            app_min = minimize_locale(app_std)
            if req_min == app_min:
                return orig

        # 4. Same-language fallback - return first same-language match
        return same_lang[0][0]

    return None



from __future__ import annotations

from generaltranslation.static._constants import VAR_IDENTIFIER
from generaltranslation.static._traverse_icu import is_gt_unindexed_select, traverse_icu


def _find_other_span(icu_string: str, node_start: int, node_end: int) -> tuple[int, int]:
    """Find the ``{content}`` span of the ``other`` option within a select node.

    Returns ``(brace_open, brace_close_exclusive)`` — the positions of the
    opening ``{`` and one past the closing ``}`` of the ``other`` option's content
    block in *icu_string*.
    """
    s = icu_string
    i = node_start

    # Skip opening '{'
    i += 1

    # Skip whitespace
    while i < node_end and s[i] in " \t\n\r":
        i += 1

    # Skip variable name
    while i < node_end and s[i] not in " \t\n\r,{}":
        i += 1

    # Skip whitespace
    while i < node_end and s[i] in " \t\n\r":
        i += 1

    # Skip comma after name
    if i < node_end and s[i] == ",":
        i += 1

    # Skip whitespace
    while i < node_end and s[i] in " \t\n\r":
        i += 1

    # Skip type name (e.g., "select")
    while i < node_end and s[i] not in " \t\n\r,{}":
        i += 1

    # Skip whitespace
    while i < node_end and s[i] in " \t\n\r":
        i += 1

    # Skip comma after type
    if i < node_end and s[i] == ",":
        i += 1

    # Now we're in the submessage area — parse selector {content} pairs
    while i < node_end:
        # Skip whitespace
        while i < node_end and s[i] in " \t\n\r":
            i += 1

        if i >= node_end or s[i] == "}":
            break

        # Read selector name
        sel_start = i
        while i < node_end and s[i] not in " \t\n\r{}":
            i += 1
        selector = s[sel_start:i]

        # Skip whitespace before '{'
        while i < node_end and s[i] in " \t\n\r":
            i += 1

        if i >= node_end or s[i] != "{":
            break

        # Find matching '}'
        brace_start = i
        depth = 1
        i += 1
        while i < node_end and depth > 0:
            c = s[i]
            if c == "{":
                depth += 1
            elif c == "}":
                depth -= 1
            elif c == "'":
                # ICU escape — skip past escaped content
                i += 1
                if i < node_end and s[i] == "'":
                    # Doubled quote ''
                    pass
                else:
                    # Escaped block until next unescaped single quote
                    while i < node_end and s[i] != "'":
                        i += 1
            i += 1
        brace_end = i  # one past closing '}'

        if selector == "other":
            return (brace_start, brace_end)

    # Fallback — should not happen for valid GT selects
    return (node_end, node_end)


def index_vars(icu_string: str) -> str:
    """Add numeric indices to unindexed ``_gt_`` selects and collapse their ``other`` content.

    Transforms ``{_gt_, select, other {value}}`` into ``{_gt_1, select, other {}}``.

    Args:
        icu_string: An ICU MessageFormat string with unindexed GT variables.

    Returns:
        The string with indexed GT variables and empty ``other`` options.
    """
    if VAR_IDENTIFIER not in icu_string:
        return icu_string

    locations: list[tuple[int, int, int, int]] = []

    def visitor(child: dict) -> None:
        node_start = child.get("start", 0)
        node_end = child.get("end", 0)
        other_start, other_end = _find_other_span(icu_string, node_start, node_end)
        locations.append((node_start, node_end, other_start, other_end))

    traverse_icu(
        icu_string,
        is_gt_unindexed_select,
        visitor,
        recurse_into_visited=False,
        include_indices=True,
    )

    result: list[str] = []
    current = 0
    for i, (start, end, other_start, other_end) in enumerate(locations):
        # Before the variable
        result.append(icu_string[current:start])
        # {_gt_ (opening brace + identifier)
        result.append(icu_string[start : start + len(VAR_IDENTIFIER) + 1])
        # Index number
        result.append(str(i + 1))
        # From after the identifier to start of other's {content}
        result.append(icu_string[start + len(VAR_IDENTIFIER) + 1 : other_start])
        # Collapsed other
        result.append("{}")
        # After other's {content} to end of select
        result.append(icu_string[other_end:end])
        current = end

    result.append(icu_string[current:])

    return "".join(result)



"""Date/time formatting using Babel.

Ports ``_formatDateTime`` from the JS core library.
"""

from __future__ import annotations

from datetime import datetime

from babel.dates import format_date, format_datetime, format_time

from generaltranslation.formatting._helpers import _resolve_babel_locale


def format_date_time(
    value: datetime | str,
    locales: str | list[str] | None = None,
    options: dict | None = None,
) -> str:
    """Format a date/time value according to the given locale and options.

    Args:
        value: A :class:`datetime` object or an ISO 8601 string.
        locales: BCP 47 locale tag(s). Defaults to ``"en"``.
        options: Formatting options (snake_case).

            - ``date_style``: ``"full"``, ``"long"``, ``"medium"``, ``"short"``.
            - ``time_style``: ``"full"``, ``"long"``, ``"medium"``, ``"short"``.

            If both are provided, ``format_datetime`` is used.
            If only ``date_style``, ``format_date`` is used.
            If only ``time_style``, ``format_time`` is used.
            If neither, ``format_datetime`` with ``"medium"`` is used.

    Returns:
        The formatted date/time string.
    """
    if options is None:
        options = {}

    if isinstance(value, str):
        value = datetime.fromisoformat(value)

    locale = _resolve_babel_locale(locales)
    date_style = options.get("date_style")
    time_style = options.get("time_style")

    if date_style and time_style:
        return (
            format_datetime(value, format=date_style, locale=locale)
            + " "
            + format_time(value, format=time_style, locale=locale)
        )

    if date_style:
        return format_date(value, format=date_style, locale=locale)

    if time_style:
        return format_time(value, format=time_style, locale=locale)

    # Default: medium datetime
    return format_datetime(value, format="medium", locale=locale)



"""Shared helpers for the formatting module.

Provides locale resolution and option mapping utilities used by
multiple formatting functions.
"""

from __future__ import annotations

from babel import Locale
from babel.core import UnknownLocaleError

from generaltranslation._settings import LIBRARY_DEFAULT_LOCALE


def _resolve_babel_locale(locales: str | list[str] | None = None) -> Locale:
    """Convert BCP 47 locale(s) to a Babel :class:`Locale` instance.

    Tries each candidate in order and falls back to ``en`` if none parse.

    Args:
        locales: A single BCP 47 tag, a list of tags, or ``None``.

    Returns:
        A :class:`babel.Locale` instance.
    """
    if locales is None:
        return Locale(LIBRARY_DEFAULT_LOCALE)
    if isinstance(locales, str):
        locales = [locales]
    for tag in locales:
        try:
            return Locale.parse(tag, sep="-")
        except (UnknownLocaleError, ValueError):
            continue
    return Locale(LIBRARY_DEFAULT_LOCALE)


def _get_language_code(locales: str | list[str] | None = None) -> str:
    """Extract the language code from a locale specification.

    Args:
        locales: A single BCP 47 tag, a list of tags, or ``None``.

    Returns:
        A 2- or 3-letter language code string.
    """
    return _resolve_babel_locale(locales).language



"""Locale utilities: validation, properties, names, direction, emoji.

This subpackage provides BCP 47 locale handling functions, ported from
the JS ``generaltranslation`` core library.  All public functions are
re-exported here for convenience.
"""

from generaltranslation.locales._custom_locale_mapping import (
    get_custom_property,
    should_use_canonical_locale,
)
from generaltranslation.locales._determine_locale import determine_locale
from generaltranslation.locales._get_locale_direction import get_locale_direction
from generaltranslation.locales._get_locale_emoji import get_locale_emoji
from generaltranslation.locales._get_locale_name import get_locale_name
from generaltranslation.locales._get_locale_properties import (
    get_locale_properties,
)
from generaltranslation.locales._get_plural_form import get_plural_form
from generaltranslation.locales._get_region_properties import get_region_properties
from generaltranslation.locales._is_same_dialect import is_same_dialect
from generaltranslation.locales._is_same_language import is_same_language
from generaltranslation.locales._is_superset_locale import is_superset_locale
from generaltranslation.locales._is_valid_locale import (
    is_valid_locale,
    standardize_locale,
)
from generaltranslation.locales._requires_translation import requires_translation
from generaltranslation.locales._resolve_locale import (
    resolve_alias_locale,
    resolve_canonical_locale,
)
from generaltranslation.locales._types import (
    PLURAL_FORMS,
    CustomMapping,
    CustomRegionMapping,
    LocaleProperties,
    PluralType,
    is_accepted_plural_form,
)

__all__ = [
    # Types & constants
    "LocaleProperties",
    "CustomMapping",
    "CustomRegionMapping",
    "PluralType",
    "PLURAL_FORMS",
    "is_accepted_plural_form",
    # Custom mapping helpers
    "get_custom_property",
    "should_use_canonical_locale",
    # Validation & standardisation
    "is_valid_locale",
    "standardize_locale",
    # Metadata
    "get_locale_properties",
    "get_locale_name",
    "get_locale_emoji",
    "get_locale_direction",
    "get_region_properties",
    # Resolution & matching
    "determine_locale",
    "resolve_canonical_locale",
    "resolve_alias_locale",
    # Comparison
    "is_same_language",
    "is_same_dialect",
    "is_superset_locale",
    # Translation logic
    "requires_translation",
    # Plurals
    "get_plural_form",
]



"""Dialect-level locale comparison.

Ports ``isSameDialect.ts`` from the JS core library.
Two locales are the "same dialect" when their language matches and
neither has a region or script that conflicts with the other.
"""

from __future__ import annotations

from babel import Locale


def is_same_dialect(*locales: str | list[str]) -> bool:
    """Return ``True`` if all given locales are the same dialect.

    Two locales are considered the same dialect when:

    - Their language codes are identical.
    - If both specify a region, the regions match.
    - If both specify a script, the scripts match.

    An unspecified (empty) region or script is treated as compatible
    with any value, so ``"en-US"`` and ``"en"`` are the same dialect,
    but ``"en-US"`` and ``"en-GB"`` are not.

    All pairwise combinations are checked (not just consecutive pairs).

    Args:
        *locales: Two or more BCP 47 locale tags (or lists thereof).

    Returns:
        ``True`` if every pair of locales is the same dialect,
        ``False`` otherwise (including on parse errors).

    Examples:
        >>> is_same_dialect("en-US", "en")
        True
        >>> is_same_dialect("en-US", "en-GB")
        False
        >>> is_same_dialect("zh-Hans", "zh-Hant")
        False
    """
    # Flatten args
    flat: list[str] = []
    for arg in locales:
        if isinstance(arg, list):
            flat.extend(arg)
        else:
            flat.append(arg)

    if len(flat) < 2:
        return False

    try:
        from generaltranslation.locales._is_valid_locale import standardize_locale

        parsed_list = []
        for loc in flat:
            std = standardize_locale(loc)
            parsed = Locale.parse(std, sep="-")
            parsed_list.append(parsed)

        # Check all pairs
        for i in range(len(parsed_list)):
            for j in range(i + 1, len(parsed_list)):
                a = parsed_list[i]
                b = parsed_list[j]

                if a.language != b.language:
                    return False

                if a.territory and b.territory and a.territory != b.territory:
                    return False

                if a.script and b.script and a.script != b.script:
                    return False

        return True
    except Exception:
        return False



"""Batch utilities for splitting and processing items."""

from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from typing import Any, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def create_batches(items: list[T], batch_size: int = 100) -> list[list[T]]:
    """Split *items* into batches of at most *batch_size*."""
    batches: list[list[T]] = []
    for i in range(0, len(items), batch_size):
        batches.append(items[i : i + batch_size])
    return batches


async def process_batches(
    items: list[Any],
    processor: Callable[[list[Any]], Awaitable[list[Any]]],
    *,
    batch_size: int = 100,
    parallel: bool = True,
) -> dict[str, Any]:
    """Process *items* in batches using *processor*.

    Returns a dict with keys ``data``, ``count``, ``batch_count``.
    """
    if not items:
        return {"data": [], "count": 0, "batch_count": 0}

    batches = create_batches(items, batch_size)
    all_items: list[Any] = []

    if parallel:
        results = await asyncio.gather(*(processor(batch) for batch in batches))
        for result in results:
            if result:
                all_items.extend(result)
    else:
        for batch in batches:
            result = await processor(batch)
            if result:
                all_items.extend(result)

    return {"data": all_items, "count": len(all_items), "batch_count": len(batches)}



"""Region metadata lookup.

Ports ``getRegionProperties.ts`` from the JS core library.
Returns the display name and flag emoji for an ISO 3166-1 alpha-2 or
UN M.49 region code.
"""

from __future__ import annotations

from babel import Locale

from generaltranslation._settings import LIBRARY_DEFAULT_LOCALE
from generaltranslation.locales._get_locale_emoji import DEFAULT_EMOJI, EMOJIS
from generaltranslation.locales._types import CustomRegionMapping


def get_region_properties(
    region: str,
    default_locale: str | None = LIBRARY_DEFAULT_LOCALE,
    custom_mapping: CustomRegionMapping | None = None,
) -> dict[str, str]:
    """Return metadata for a region code.

    Looks up a region by its ISO 3166-1 alpha-2 (e.g. ``"US"``) or
    UN M.49 (e.g. ``"419"``) code and returns its localised display
    name and flag emoji.

    Custom mapping entries override the default name and/or emoji.

    Args:
        region: The region code to look up.
        default_locale: The locale in which to render the region name
            (defaults to ``"en"``).
        custom_mapping: Optional mapping of region codes to custom
            ``{"name": ..., "emoji": ..., "locale": ...}`` overrides.

    Returns:
        A dict with keys ``"code"``, ``"name"``, and ``"emoji"``.
        May also include ``"locale"`` if provided by *custom_mapping*.

    Examples:
        >>> get_region_properties("US")
        {'code': 'US', 'name': 'United States', 'emoji': '🇺🇸'}
        >>> get_region_properties("US", default_locale="fr")
        {'code': 'US', 'name': 'États-Unis', 'emoji': '🇺🇸'}
    """
    result: dict[str, str] = {"code": region}

    # Get emoji from EMOJIS dict
    result["emoji"] = EMOJIS.get(region, DEFAULT_EMOJI)

    # Get region name from Babel
    try:
        display_locale = Locale.parse(default_locale, sep="-")
        name = display_locale.territories.get(region, "")
        result["name"] = name
    except Exception:
        result["name"] = ""

    # Apply custom mapping overrides
    if custom_mapping is not None and region in custom_mapping:
        custom = custom_mapping[region]
        if "name" in custom:
            result["name"] = custom["name"]
        if "emoji" in custom:
            result["emoji"] = custom["emoji"]
        if "locale" in custom:
            result["locale"] = custom["locale"]

    return result



"""Retrieve comprehensive metadata for a BCP 47 locale.

Ports ``getLocaleProperties.ts`` from the JS core library.
Uses ``babel`` for parsing, maximisation/minimisation, and display
name resolution instead of the browser ``Intl`` APIs.
"""

from __future__ import annotations

from babel import Locale
from babel.core import get_global

from generaltranslation._settings import LIBRARY_DEFAULT_LOCALE
from generaltranslation.locales._types import CustomMapping, LocaleProperties

_likely_subtags: dict[str, str] = dict(get_global("likely_subtags"))


def _create_custom_locale_properties(
    locale_variants: list[str],
    custom_mapping: CustomMapping | None = None,
) -> dict[str, str] | None:
    """Merge custom property overrides from the mapping."""
    if custom_mapping is None:
        return None

    result: dict[str, str] = {}
    for variant in reversed(locale_variants):
        entry = custom_mapping.get(variant)
        if entry is not None:
            if isinstance(entry, str):
                result["name"] = entry
            elif isinstance(entry, dict):
                result.update(entry)

    return result if result else None


def _apply_cjk_formatting(text: str, display_lang: str) -> str:
    """Apply CJK-specific formatting to display names."""
    if display_lang == "zh":
        text = text.replace(" (", "\uff08").replace(")", "\uff09")
        text = text.replace(", ", "\uff0c")
    elif display_lang == "ja":
        text = text.replace(", ", "\u3001")
    return text


def _get_compound_name(
    locale_tag: str,
    display_locale: Locale,
    parsed: Locale,
) -> str:
    """Get compound CLDR name (e.g. 'Austrian German') if available,
    otherwise fall back to get_display_name."""
    underscore_tag = parsed.language
    if parsed.script:
        underscore_tag += f"_{parsed.script}"
    if parsed.territory:
        underscore_tag += f"_{parsed.territory}"

    compound = display_locale.languages.get(underscore_tag)
    if compound:
        return compound

    name = parsed.get_display_name(str(display_locale))
    if name:
        name = _apply_cjk_formatting(name, display_locale.language)
        return name
    return locale_tag


def _build_component_name(
    language: str,
    script: str | None,
    territory: str | None,
    display_locale: Locale,
) -> str:
    """Build 'Language (Script, Territory)' name with locale-specific formatting."""
    lang_name = display_locale.languages.get(language, language)
    details = []
    if script:
        details.append(display_locale.scripts.get(script, script))
    if territory:
        details.append(display_locale.territories.get(territory, territory))

    if not details:
        return lang_name

    joined = ", ".join(details)
    result = f"{lang_name} ({joined})"

    # Apply locale-specific formatting for CJK
    display_lang = display_locale.language
    if display_lang == "zh":
        # Chinese: fullwidth brackets and comma
        fw_comma = "\uff0c"
        result = f"{lang_name}\uff08{fw_comma.join(details)}\uff09"
    elif display_lang == "ja":
        # Japanese: ASCII parens, ideographic comma
        jp_comma = "\u3001"
        result = f"{lang_name} ({jp_comma.join(details)})"

    return result


def get_locale_properties(
    locale: str,
    default_locale: str | None = LIBRARY_DEFAULT_LOCALE,
    custom_mapping: CustomMapping | None = None,
) -> LocaleProperties:
    """Return a :class:`LocaleProperties` for *locale*."""
    from generaltranslation.locales._custom_locale_mapping import (
        should_use_canonical_locale,
    )
    from generaltranslation.locales._get_locale_emoji import get_locale_emoji
    from generaltranslation.locales._is_valid_locale import standardize_locale
    from generaltranslation.locales.utils import minimize_locale

    original_locale = locale

    # Resolve canonical if needed
    canonical_code = locale
    if custom_mapping is not None and should_use_canonical_locale(locale, custom_mapping):
        entry = custom_mapping[locale]
        if isinstance(entry, dict) and "code" in entry:
            canonical_code = entry["code"]

    std_locale = standardize_locale(canonical_code)

    try:
        parsed = Locale.parse(std_locale, sep="-")
    except Exception:
        return LocaleProperties(
            code=original_locale,
            name=original_locale,
            native_name=original_locale,
        )

    # Get display locale
    try:
        display_locale = Locale.parse(default_locale, sep="-")
    except Exception:
        display_locale = Locale("en")

    # Components from parsed
    language_code = parsed.language
    input_region = parsed.territory or ""
    input_script = parsed.script or ""

    # Maximize using likely subtags
    max_script = input_script
    max_territory = input_region

    candidates = []
    if input_script and input_region:
        candidates.append(f"{language_code}_{input_script}_{input_region}")
    if input_script:
        candidates.append(f"{language_code}_{input_script}")
    if input_region:
        candidates.append(f"{language_code}_{input_region}")
    candidates.append(language_code)

    for candidate in candidates:
        if candidate in _likely_subtags:
            parts = _likely_subtags[candidate].split("_")
            if not max_script and len(parts) > 1:
                max_script = parts[1]
            if not max_territory and len(parts) > 2:
                max_territory = parts[2]
            break

    # Fill remaining from base language
    if not max_script or not max_territory:
        if language_code in _likely_subtags:
            parts = _likely_subtags[language_code].split("_")
            if not max_script and len(parts) > 1:
                max_script = parts[1]
            if not max_territory and len(parts) > 2:
                max_territory = parts[2]

    # Build maximized/minimized codes
    maximized_parts = [language_code]
    if max_script:
        maximized_parts.append(max_script)
    if max_territory:
        maximized_parts.append(max_territory)
    maximized_code = "-".join(maximized_parts)

    minimized_code = minimize_locale(std_locale)

    # Parse minimized for display name lookup
    try:
        minimized_parsed = Locale.parse(minimized_code, sep="-")
    except Exception:
        minimized_parsed = parsed

    # Use maximized region/script as defaults for properties
    region_code = input_region or max_territory
    script_code = input_script or max_script

    # --- Display names in default_locale ---
    name = _get_compound_name(std_locale, display_locale, parsed)
    language_name = display_locale.languages.get(language_code, language_code)
    region_name = display_locale.territories.get(region_code, "") if region_code else ""
    script_name = display_locale.scripts.get(script_code, "") if script_code else ""

    # Maximized name: always use component form
    maximized_name = _build_component_name(language_code, max_script, max_territory, display_locale)

    # Minimized name: use compound lookup
    minimized_name = _get_compound_name(minimized_code, display_locale, minimized_parsed)

    # --- Native display names ---
    # Include script in native locale for correct display (e.g. sr-Latn vs sr-Cyrl)
    native_locale_id = language_code
    if parsed.script:
        native_locale_id += f"_{parsed.script}"
    try:
        native_locale = Locale.parse(native_locale_id)
    except Exception:
        try:
            native_locale = Locale.parse(language_code)
        except Exception:
            native_locale = parsed

    native_name = _get_compound_name(std_locale, native_locale, parsed)
    native_language_name = native_locale.languages.get(language_code, language_code)
    native_region_name = native_locale.territories.get(region_code, "") if region_code else ""
    native_script_name = native_locale.scripts.get(script_code, "") if script_code else ""
    native_maximized_name = _build_component_name(language_code, max_script, max_territory, native_locale)
    native_minimized_name = _get_compound_name(minimized_code, native_locale, minimized_parsed)

    # Name with region code
    if parsed.territory:
        name_with_region_code = f"{language_name} ({parsed.territory})"
        native_name_with_region_code = f"{native_language_name} ({parsed.territory})"
    else:
        name_with_region_code = language_name
        native_name_with_region_code = native_language_name

    emoji = get_locale_emoji(std_locale, custom_mapping)

    props = LocaleProperties(
        code=original_locale,
        name=name,
        native_name=native_name,
        language_code=language_code,
        language_name=language_name,
        native_language_name=native_language_name,
        name_with_region_code=name_with_region_code,
        native_name_with_region_code=native_name_with_region_code,
        region_code=region_code,
        region_name=region_name,
        native_region_name=native_region_name,
        script_code=script_code,
        script_name=script_name,
        native_script_name=native_script_name,
        maximized_code=maximized_code,
        maximized_name=maximized_name,
        native_maximized_name=native_maximized_name,
        minimized_code=minimized_code,
        minimized_name=minimized_name,
        native_minimized_name=native_minimized_name,
        emoji=emoji,
    )

    # Apply custom overrides
    if custom_mapping is not None:
        locale_variants = [original_locale, std_locale, language_code]
        custom_props = _create_custom_locale_properties(locale_variants, custom_mapping)
        if custom_props:
            key_map = {
                "name": "name",
                "nativeName": "native_name",
                "emoji": "emoji",
                "languageCode": "language_code",
                "languageName": "language_name",
                "regionCode": "region_code",
                "regionName": "region_name",
                "scriptCode": "script_code",
                "scriptName": "script_name",
            }
            for custom_key, value in custom_props.items():
                field = key_map.get(custom_key, custom_key)
                if hasattr(props, field) and field != "code":
                    setattr(props, field, value)

    return props



"""Enqueue files for translation."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._batch import process_batches
from generaltranslation.translate._request import api_request


async def enqueue_files(
    files: list[dict[str, Any]],
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Enqueue files for translation in batches."""

    async def _process_batch(batch: list[dict[str, Any]]) -> list[Any]:
        body: dict[str, Any] = {
            "files": [
                {
                    "branchId": f.get("branch_id", f.get("branchId", "")),
                    "fileId": f.get("file_id", f.get("fileId", "")),
                    "versionId": f.get("version_id", f.get("versionId", "")),
                    "fileName": f.get("file_name", f.get("fileName", "")),
                }
                for f in batch
            ],
            "targetLocales": options.get("target_locales", options.get("targetLocales", [])),
            "sourceLocale": options.get("source_locale", options.get("sourceLocale", "")),
            "publish": options.get("publish"),
            "requireApproval": options.get("require_approval", options.get("requireApproval")),
            "modelProvider": options.get("model_provider", options.get("modelProvider")),
            "force": options.get("force"),
        }
        body = {k: v for k, v in body.items() if v is not None}
        result = await api_request(
            config,
            "/v2/project/translations/enqueue",
            body=body,
            timeout=options.get("timeout"),
        )
        return list(result.get("jobData", {}).items())

    result = await process_batches(files, _process_batch, batch_size=100)
    jobs = dict(result["data"])
    target_locales = options.get("target_locales", options.get("targetLocales", []))
    return {
        "jobData": jobs,
        "locales": target_locales,
        "message": (
            f"Successfully enqueued {result['count']} file translation jobs in {result['batch_count']} batch(es)"
        ),
    }



"""Translate and translate_many functions."""

from __future__ import annotations

from typing import Any

from generaltranslation._id import hash_source
from generaltranslation._settings import DEFAULT_RUNTIME_API_URL
from generaltranslation.translate._request import api_request


async def translate_many(
    requests: list[Any] | dict[str, Any],
    global_metadata: dict[str, Any],
    config: dict[str, Any],
    timeout: int | None = None,
) -> list[dict[str, Any]] | dict[str, dict[str, Any]]:
    """Translate multiple entries in a single API request.

    Parameters
    ----------
    requests : list or dict
        If a list, each entry is a string or ``{"source": ..., "metadata": ...}``.
        If a dict, keys are hashes and values are entries.
    global_metadata : dict
        Must contain ``target_locale`` and ``source_locale``.
    config : dict
        Translation request config.
    timeout : int, optional
        Timeout in ms.

    Returns
    -------
    list or dict
        Array of results if input was list, dict if input was dict.
    """
    is_array = isinstance(requests, list)

    hash_order: list[str] | None = [] if is_array else None
    requests_object: dict[str, dict[str, Any]] = {}

    if is_array:
        entries: list[tuple[str | None, Any]] = [(None, r) for r in requests]
    else:
        assert isinstance(requests, dict)
        entries = list(requests.items())

    for key, request in entries:
        if isinstance(request, str):
            normalized: dict[str, Any] = {"source": request}
        else:
            normalized = request

        source = normalized.get("source")
        metadata = normalized.get("metadata") or {}

        if key is not None:
            entry_hash = key
        elif metadata.get("hash"):
            entry_hash = metadata["hash"]
        else:
            entry_hash = hash_source(
                source,
                data_format=metadata.get("dataFormat", metadata.get("data_format", "STRING")),
                context=metadata.get("context"),
                id=metadata.get("id"),
                max_chars=metadata.get("maxChars", metadata.get("max_chars")),
            )

        if hash_order is not None:
            hash_order.append(entry_hash)

        requests_object[entry_hash] = {
            "source": source,
            "metadata": metadata or None,
        }

    # Build request body using camelCase keys to match JS API
    body = {
        "requests": requests_object,
        "targetLocale": global_metadata.get("target_locale", global_metadata.get("targetLocale", "")),
        "sourceLocale": global_metadata.get("source_locale", global_metadata.get("sourceLocale", "")),
        "metadata": global_metadata,
    }

    # Use runtime API URL
    request_config = {
        **config,
        "base_url": config.get("base_url") or DEFAULT_RUNTIME_API_URL,
    }

    response = await api_request(
        request_config,
        "/v2/translate",
        body=body,
        timeout=timeout,
        retry_policy="none",
    )

    if hash_order is not None:
        return [
            response.get(h, {"success": False, "error": "No translation returned", "code": 500}) for h in hash_order
        ]

    return response



"""Query source file endpoint."""

from __future__ import annotations

from typing import Any
from urllib.parse import quote, urlencode

from generaltranslation.translate._request import api_request


async def query_source_file(
    query: dict[str, Any],
    options: dict[str, Any],
    config: dict[str, Any],
) -> dict[str, Any]:
    """Get source file and translation information for a given file ID."""
    file_id = query.get("file_id", query.get("fileId", ""))
    branch_id = query.get("branch_id", query.get("branchId"))
    version_id = query.get("version_id", query.get("versionId"))

    params: dict[str, str] = {}
    if branch_id:
        params["branchId"] = branch_id
    if version_id:
        params["versionId"] = version_id

    qs = urlencode(params) if params else ""
    encoded_file_id = quote(file_id, safe="!'()*-._~")
    endpoint = f"/v2/project/translations/files/status/{encoded_file_id}"
    if qs:
        endpoint = f"{endpoint}?{qs}"

    return await api_request(
        config,
        endpoint,
        method="GET",
        timeout=options.get("timeout"),
    )



"""Error types for General Translation."""

from generaltranslation.errors._api_error import ApiError
from generaltranslation.errors._messages import (
    GT_ERROR_PREFIX,
    INVALID_AUTH_ERROR,
    api_error_message,
    create_invalid_cutoff_style_error,
    invalid_locale_error,
    invalid_locales_error,
    no_api_key_error,
    no_project_id_error,
    no_source_locale_error,
    no_target_locale_error,
    translation_request_failed_error,
    translation_timeout_error,
)

__all__ = [
    "ApiError",
    "GT_ERROR_PREFIX",
    "INVALID_AUTH_ERROR",
    "api_error_message",
    "create_invalid_cutoff_style_error",
    "invalid_locale_error",
    "invalid_locales_error",
    "no_api_key_error",
    "no_project_id_error",
    "no_source_locale_error",
    "no_target_locale_error",
    "translation_request_failed_error",
    "translation_timeout_error",
]



"""Submit user edit diffs endpoint."""

from __future__ import annotations

from typing import Any

from generaltranslation.translate._batch import process_batches
from generaltranslation.translate._request import api_request


async def submit_user_edit_diffs(
    payload: dict[str, Any],
    config: dict[str, Any],
    options: dict[str, Any] | None = None,
) -> dict[str, bool]:
    """Submit user edit diffs so the service can learn/persist user-intended rules."""
    options = options or {}
    diffs = payload.get("diffs", [])

    async def _process_batch(batch: list[Any]) -> list[Any]:
        await api_request(
            config,
            "/v2/project/files/diffs",
            body={"diffs": batch},
            timeout=options.get("timeout"),
        )
        return [{"success": True}]

    await process_batches(diffs, _process_batch, batch_size=100)
    return {"success": True}



"""Translation API communication layer."""

from generaltranslation.translate._batch import create_batches, process_batches
from generaltranslation.translate._check_job_status import check_job_status
from generaltranslation.translate._create_branch import create_branch
from generaltranslation.translate._download_file_batch import download_file_batch
from generaltranslation.translate._enqueue_files import enqueue_files
from generaltranslation.translate._get_orphaned_files import get_orphaned_files
from generaltranslation.translate._get_project_data import get_project_data
from generaltranslation.translate._headers import generate_request_headers
from generaltranslation.translate._process_file_moves import process_file_moves
from generaltranslation.translate._query_branch_data import query_branch_data
from generaltranslation.translate._query_file_data import query_file_data
from generaltranslation.translate._query_source_file import query_source_file
from generaltranslation.translate._request import api_request
from generaltranslation.translate._setup_project import setup_project
from generaltranslation.translate._submit_user_edit_diffs import submit_user_edit_diffs
from generaltranslation.translate._translate import translate_many
from generaltranslation.translate._upload_source_files import upload_source_files
from generaltranslation.translate._upload_translations import upload_translations

__all__ = [
    "api_request",
    "check_job_status",
    "create_batches",
    "create_branch",
    "download_file_batch",
    "enqueue_files",
    "generate_request_headers",
    "get_orphaned_files",
    "get_project_data",
    "process_batches",
    "process_file_moves",
    "query_branch_data",
    "query_file_data",
    "query_source_file",
    "setup_project",
    "submit_user_edit_diffs",
    "translate_many",
    "upload_source_files",
    "upload_translations",
]
